from __future__ import division
import re
import math
import logging
import os

from avocado.core import exceptions
from avocado.utils import process

from virttest import kernel_interface


# Returns total memory in kb
def read_from_meminfo(key, session=None):
    """
    wrapper to return value from /proc/meminfo using key

    :param key: filter based on the key
    :param session: ShellSession Object of remote host / guest
    :return: value mapped to the key of type int
    """
    func = process.getoutput
    if session:
        func = session.cmd_output
    meminfo = func('grep %s /proc/meminfo' % key)
    return int(re.search(r'\d+', meminfo).group(0))


def memtotal(session=None):
    """
    Method to get the memtotal from /proc/meminfo

    :param session: ShellSession Object of remote host / guest
    """
    return read_from_meminfo('MemTotal', session=session)


def freememtotal(session=None):
    """
    Method to get the freememtotal from /proc/meminfo

    :param session: ShellSession Object of remote host / guest
    """
    return read_from_meminfo('MemFree', session=session)


def rounded_memtotal(session=None):
    """
    Method to get total of all physical mem, in kbytes

    :param session: ShellSession Object of remote host / guest
    """
    usable_kbytes = memtotal(session=session)
    # usable_kbytes is system's usable DRAM in kbytes,
    #   as reported by memtotal() from device /proc/meminfo memtotal
    #   after Linux deducts 1.5% to 5.1% for system table overhead
    # Undo the unknown actual deduction by rounding up
    #   to next small multiple of a big power-of-two
    #   eg  12GB - 5.1% gets rounded back up to 12GB
    mindeduct = 0.015  # 1.5 percent
    maxdeduct = 0.055  # 5.5 percent
    # deduction range 1.5% .. 5.5% supports physical mem sizes
    #    6GB .. 12GB in steps of .5GB
    #   12GB .. 24GB in steps of 1 GB
    #   24GB .. 48GB in steps of 2 GB ...
    # Finer granularity in physical mem sizes would require
    #   tighter spread between min and max possible deductions

    # increase mem size by at least min deduction, without rounding
    min_kbytes = int(usable_kbytes / (1.0 - mindeduct))
    # increase mem size further by 2**n rounding, by 0..roundKb or more
    round_kbytes = int(usable_kbytes / (1.0 - maxdeduct)) - min_kbytes
    # find least binary roundup 2**n that covers worst-cast roundKb
    mod2n = 1 << int(math.ceil(math.log(round_kbytes, 2)))
    # have round_kbytes <= mod2n < round_kbytes*2
    # round min_kbytes up to next multiple of mod2n
    phys_kbytes = min_kbytes + mod2n - 1
    phys_kbytes = phys_kbytes - (phys_kbytes % mod2n)  # clear low bits
    return phys_kbytes


def numa_nodes(session=None):
    """
    Method to get total no of numa nodes

    :param session: ShellSession Object of remote host / guest
    """
    func = process.getoutput
    if session:
        func = session.cmd_output
    base_path = "/sys/devices/system/node"
    node_avail = func("ls %s | grep 'node'" % base_path).split()
    node_paths = [os.path.join(base_path, each_node) for each_node in node_avail]
    nodes = [int(re.sub(r'.*node(\d+)', r'\1', x)) for x in node_paths]
    return (sorted(nodes))


def node_size(session=None):
    """
    Method to get node size

    :param session: ShellSession Object of remote host / guest
    """
    nodes = max(len(numa_nodes(session=session)), 1)
    return ((memtotal() * 1024) // nodes)


def get_huge_page_size(session=None):
    """
    Method to get huge page size

    :param session: ShellSession Object of remote host / guest
    """
    return read_from_meminfo('Hugepagesize', session=session)


def get_num_huge_pages(session=None):
    """
    Method to get total no of hugepages

    :param session: ShellSession Object of remote host / guest
    """
    return read_from_meminfo('HugePages_Total', session=session)


def get_num_huge_pages_free(session=None):
    """
    Method to get free hugepages available

    :param session: ShellSession Object of remote host / guest
    """
    return read_from_meminfo('HugePages_Free', session=session)


def get_num_huge_pages_rsvd(session=None):
    """
    Method to get reserved hugepage pages

    :param session: ShellSession Object of remote host / guest
    """
    return read_from_meminfo('HugePages_Rsvd', session=session)


def get_num_huge_pages_surp(session=None):
    """
    Method to get surplus hugepage pages
    :param session: ShellSession Object of remote host / guest
    """
    return read_from_meminfo('HugePages_Surp', session=session)


def get_num_anon_huge_pages(pid=0, session=None):
    """
    Method to get total no of anon hugepages

    :param pid: pid of the specific process
    :param session: ShellSession Object of remote host / guest
    """
    if int(pid) > 1:
        # get AnonHugePages usage of specified process
        return read_from_smaps(pid, 'AnonHugePages', session=session)
    else:
        # invalid pid, so return AnonHugePages of the host
        return read_from_meminfo('AnonHugePages', session=session)


def get_transparent_hugepage(session=None, regex="[]"):
    """
    Method to get total no of transparent hugepage

    :param regex: regex used to hightlight the selected value
    :param session: ShellSession Object of remote host / guest
    """
    UPSTREAM_THP_PATH = "/sys/kernel/mm/transparent_hugepage"
    RH_THP_PATH = "/sys/kernel/mm/redhat_transparent_hugepage"
    if os.path.isdir(UPSTREAM_THP_PATH):
        thp_path = UPSTREAM_THP_PATH
    elif os.path.isdir(RH_THP_PATH):
        thp_path = RH_THP_PATH
    else:
        raise exceptions.TestFail("transparent hugepage Not supported")
    thp = kernel_interface.SysFS(os.path.join(thp_path, 'enabled'),
                                 session=session)
    return thp.sys_fs_value.strip(regex)


def set_num_huge_pages(num, session=None):
    """
    Method to set no of transparent hugepages

    :param num: value to be set for THP
    :param session: ShellSession Object of remote host / guest
    """
    func = process.system
    if session:
        func = session.cmd_status
    return func('/sbin/sysctl vm.nr_hugepages=%d' % num) == 0


def set_transparent_hugepage(sflag, session=None):
    """
    Method to set THP parameter

    :param sflag:  only can be set always, madvise or never.
    :param session: ShellSession Object of remote host / guest
    """
    flags = ['always', 'madvise', 'never']
    if sflag not in flags:
        raise exceptions.TestFail("specify wrong parameter")
    UPSTREAM_THP_PATH = "/sys/kernel/mm/transparent_hugepage"
    RH_THP_PATH = "/sys/kernel/mm/redhat_transparent_hugepage"
    if os.path.isdir(UPSTREAM_THP_PATH):
        thp_path = UPSTREAM_THP_PATH
    elif os.path.isdir(RH_THP_PATH):
        thp_path = RH_THP_PATH
    else:
        raise exceptions.TestFail("transparent hugepage Not supported")
    thp = kernel_interface.SysFS(os.path.join(thp_path, 'enabled'),
                                 session=session)
    thp.sys_fs_value = sflag
    if sflag not in thp.sys_fs_value:
        raise exceptions.TestFail("setting transparent_hugepage failed")


def drop_caches(session=None):
    """
    Method to write back all dirty pages to disk and clears all the caches

    :param session: ShellSession Object of remote host / guest
    """
    func = process.getoutput
    if session:
        func = session.cmd_output
    func("sync")
    # We ignore failures here as this will fail on 2.6.11 kernels.
    drop_caches = kernel_interface.ProcFS("/proc/sys/vm/drop_caches",
                                          session=session)
    drop_caches.proc_fs_value = 3


def read_from_vmstat(key, session=None):
    """
    Get specific item value from vmstat

    :param key: The item you want to check from vmstat
    :type key: String
    :param session: ShellSession Object of remote host / guest
    :return: The value of the item
    :rtype: int
    """
    func = process.getoutput
    if session:
        func = session.cmd_output
    vmstat_info = func("cat /proc/vmstat")
    return int(re.findall("%s\s+(\d+)" % key, vmstat_info)[0])


def read_from_smaps(pid, key, session=None):
    """
    Get specific item value from the smaps of a process include all sections.

    :param pid: Process id
    :type pid: String
    :param key: The item you want to check from smaps
    :type key: String
    :param session: ShellSession Object of remote host / guest
    :return: The value of the item in kb
    :rtype: int
    """
    func = process.getoutput
    if session:
        func = session.cmd_output
    smaps_info = func('grep %s /proc/%s/smaps' % (key, pid))

    memory_size = 0
    for each_number in re.findall("%s:\s+(\d+)" % key, smaps_info):
        memory_size += int(each_number)

    return memory_size


def read_from_numastat(pid, key, session=None):
    """
    Get the process numastat from numastat output.

    :param pid: pid of the specific process
    :param key: filter based on the key
    :param session: ShellSession Object of remote host / guest
    """
    func = process.getoutput
    if session:
        func = session.cmd_output
    cmd = "numastat %s" % pid
    numa_mem = func(cmd).strip()
    mem_line = re.findall(r"^%s.*" % key, numa_mem, re.M)[0]
    return re.findall(r"(\d+.\d+)", mem_line)


def read_from_numa_maps(pid, key, session=None):
    """
    Get the process numa related info from numa_maps. This function
    only use to get the numbers like anon=1.

    :param pid: Process id
    :type pid: String
    :param key: The item you want to check from numa_maps
    :type key: String
    :param session: ShellSession Object of remote host / guest
    :return: A dict using the address as the keys
    :rtype: dict
    """
    func = process.getoutput
    if session:
        func = session.cmd_output
    numa_map_info = func("cat /proc/%s/numa_maps" % pid)

    numa_maps_dict = {}
    numa_pattern = r"(^[\dabcdfe]+)\s+.*%s[=:](\d+)" % key
    for address, number in re.findall(numa_pattern, numa_map_info, re.M):
        numa_maps_dict[address] = number

    return numa_maps_dict


def get_buddy_info(chunk_sizes, nodes="all", zones="all", session=None):
    """
    Get the fragement status of the host. It use the same method
    to get the page size in buddyinfo.
    2^chunk_size * page_size
    The chunk_sizes can be string make up by all orders that you want to check
    splited with blank or a mathematical expression with '>', '<' or '='.
    For example:
    The input of chunk_size could be: "0 2 4"
    And the return  will be: {'0': 3, '2': 286, '4': 687}
    if you are using expression: ">=9"
    the return will be: {'9': 63, '10': 225}

    :param chunk_size: The order number shows in buddyinfo. This is not
                       the real page size.
    :type chunk_size: string
    :param nodes: The numa node that you want to check. Default value is all
    :type nodes: string
    :param zones: The memory zone that you want to check. Default value is all
    :type zones: string
    :param session: ShellSession Object of remote host / guest
    :return: A dict using the chunk_size as the keys
    :rtype: dict
    """
    func = process.getoutput
    if session:
        func = session.cmd_output
    buddy_info_content = func("cat /proc/buddyinfo")

    re_buddyinfo = "Node\s+"
    if nodes == "all":
        re_buddyinfo += "(\d+)"
    else:
        re_buddyinfo += "(%s)" % "|".join(nodes.split())

    if not re.findall(re_buddyinfo, buddy_info_content):
        logging.warn("Can not find Nodes %s" % nodes)
        return None
    re_buddyinfo += ".*?zone\s+"
    if zones == "all":
        re_buddyinfo += "(\w+)"
    else:
        re_buddyinfo += "(%s)" % "|".join(zones.split())
    if not re.findall(re_buddyinfo, buddy_info_content):
        logging.warn("Can not find zones %s" % zones)
        return None
    re_buddyinfo += "\s+([\s\d]+)"

    buddy_list = re.findall(re_buddyinfo, buddy_info_content)

    if re.findall("[<>=]", chunk_sizes) and buddy_list:
        size_list = list(range(len(buddy_list[-1][-1].strip().split())))
        chunk_sizes = [str(_) for _ in size_list if eval("%s %s" % (_,
                                                                    chunk_sizes))]

        chunk_sizes = ' '.join(chunk_sizes)

    buddyinfo_dict = {}
    for chunk_size in chunk_sizes.split():
        buddyinfo_dict[chunk_size] = 0
        for _, _, chunk_info in buddy_list:
            chunk_info = chunk_info.strip().split()[int(chunk_size)]
            buddyinfo_dict[chunk_size] += int(chunk_info)

    return buddyinfo_dict


def getpagesize(session=None):
    """
    Get system page size

    :param session: ShellSession Object of VM/remote host
    :return: pagesize in kB
    """
    func = process.getoutput
    if session:
        func = session.cmd_output
    return int(func('getconf PAGE_SIZE')) // 1024



#!/usr/bin/python
"""
:author Amos Kong <akong@redhat.com>

"""
import os
import sys

if len(sys.argv) < 4:
    print(""" netperf agent usage:
    %s [session_number] [netperf_path] [netperf_parameters_str]

    $session_number: number of client sessions
    $netperf_path: client path
    $netperf_parameter_str: netperf parameters string""" % sys.argv[0])
    sys.exit()

n = int(sys.argv[1])
path = sys.argv[2]
params = " ".join(sys.argv[3:])

for i in range(n - 1):
    os.system("%s %s &" % (path, params))
os.system("%s %s" % (path, params))



import os
import random
import time
import threading
import logging

from avocado.utils import cpu

from virttest import virsh
from virttest import cpu as cpuutil
from virttest import utils_net
from virttest.utils_test import libvirt
from virttest.libvirt_xml.devices.disk import Disk


class VMStressEvents():

    def __init__(self, params, env):
        """
        :param params: test param
        """
        self.host_cpu_list = cpu.online_list() if hasattr(cpu, 'online_list') else cpu.cpu_online_list()
        self.iterations = int(params.get("stress_itrs", 1))
        self.host_iterations = int(params.get("host_event_itrs", 10))
        self.event_sleep_time = int(params.get("event_sleep_time", 10))
        self.itr_sleep_time = int(params.get("itr_sleep_time", 10))
        self.ignore_status = params.get("ignore_status", "no") == "yes"
        self.vms = env.get_all_vms()
        self.params = params
        self.host_events = params.get("host_stress_events", "")
        if self.host_events:
            self.host_events = self.host_events.split(',')
        else:
            self.host_events = []
        self.threads = []

    def run_threads(self):
        for vm in self.vms:
            vm_params = self.params.object_params(vm.name)
            events = vm_params.get("stress_events", "reboot")
            if events:
                events = events.split(',')
            else:
                events = []
            for event in events:
                self.threads.append(threading.Thread(
                    target=self.vm_stress_events, args=(event, vm, vm_params)))
        for event in self.host_events:
            self.threads.append(threading.Thread(target=self.host_stress_event, args=(event,)))
        for thread in self.threads:
            thread.start()

    def wait_for_threads(self):
        for thread in self.threads:
            thread.join()

    def vm_stress_events(self, event, vm, params):
        """
        Stress events

        :param event: event name
        :param vm: vm object
        """
        current_vcpu = int(params.get("smp", 2))
        max_vcpu = int(params.get("vcpu_maxcpus", 2))
        iface_num = params.get("iface_num", '1')
        iface_type = params.get("iface_type", "network")
        iface_model = params.get("iface_model", "virtio")
        iface_source = eval(params.get("iface_source", "{'network':'default'}"))
        attach_option = params.get("attach_option", "")
        detach_option = params.get("detach_option", "")
        disk_size = params.get("virt_disk_device_size", "1")
        disk_type = params.get("disk_type", "file")
        disk_device = params.get("disk_device", "disk")
        disk_format = params.get("disk_format", "qcow2")
        device_target = params.get("virt_disk_device_target", "vda").split()
        path = params.get("path", "")
        device_source_names = params.get("virt_disk_device_source", "").split()
        disk_driver = params.get("driver_name", "qemu")
        self.ignore_status = params.get("ignore_status", "no") == "yes"
        dargs = {'ignore_status': True, 'debug': True}
        for itr in range(self.iterations):
            if "vcpupin" in event:
                for vcpu in range(current_vcpu):
                    result = virsh.vcpupin(vm.name, vcpu,
                                           random.choice(self.host_cpu_list),
                                           **dargs)
                    if not self.ignore_status:
                        libvirt.check_exit_status(result)
            elif "emulatorpin" in event:
                result = virsh.emulatorpin(vm.name,
                                           random.choice(self.host_cpu_list),
                                           **dargs)
                if not self.ignore_status:
                    libvirt.check_exit_status(result)
            elif "suspend" in event:
                result = virsh.suspend(vm.name, **dargs)
                if not self.ignore_status:
                    libvirt.check_exit_status(result)
                time.sleep(self.event_sleep_time)
                result = virsh.resume(vm.name, **dargs)
                if not self.ignore_status:
                    libvirt.check_exit_status(result)
            elif "cpuhotplug" in event:
                result = virsh.setvcpus(vm.name, max_vcpu, "--live",
                                        **dargs)
                if not self.ignore_status:
                    libvirt.check_exit_status(result)
                    exp_vcpu = {'max_config': max_vcpu,
                                'max_live': max_vcpu,
                                'cur_config': current_vcpu,
                                'cur_live': max_vcpu,
                                'guest_live': max_vcpu}
                    cpuutil.check_vcpu_value(
                        vm, exp_vcpu, option="--live")
                time.sleep(self.event_sleep_time)
                result = virsh.setvcpus(vm.name, current_vcpu, "--live",
                                        **dargs)
                if not self.ignore_status:
                    libvirt.check_exit_status(result)
                    exp_vcpu = {'max_config': max_vcpu,
                                'max_live': max_vcpu,
                                'cur_config': current_vcpu,
                                'cur_live': current_vcpu,
                                'guest_live': current_vcpu}
                    cpuutil.check_vcpu_value(
                        vm, exp_vcpu, option="--live")
            elif "reboot" in event:
                vm.reboot()
            elif "nethotplug" in event:
                for iface_num in range(int(iface_num)):
                    logging.debug("Try to attach interface %d" % iface_num)
                    mac = utils_net.generate_mac_address_simple()
                    options = ("%s %s --model %s --mac %s %s" %
                               (iface_type, iface_source['network'],
                                iface_model, mac, attach_option))
                    logging.debug("VM name: %s , Options for Network attach: %s", vm.name, options)
                    ret = virsh.attach_interface(vm.name, options,
                                                 ignore_status=True)
                    time.sleep(self.event_sleep_time)
                    if not self.ignore_status:
                        libvirt.check_exit_status(ret)
                    if detach_option:
                        options = ("--type %s --mac %s %s" %
                                   (iface_type, mac, detach_option))
                        logging.debug("VM name: %s , Options for Network detach: %s", vm.name, options)
                        ret = virsh.detach_interface(vm.name, options,
                                                     ignore_status=True)
                        if not self.ignore_status:
                            libvirt.check_exit_status(ret)
            elif "diskhotplug" in event:
                for disk_num in range(len(device_source_names)):
                    disk = {}
                    disk_attach_error = False
                    disk_name = os.path.join(path, vm.name, device_source_names[disk_num])
                    device_source = libvirt.create_local_disk(disk_type, disk_name, disk_size, disk_format=disk_format)
                    disk.update({"format": disk_format,
                                 "source": device_source})
                    disk_xml = Disk(disk_type)
                    disk_xml.device = disk_device
                    disk_xml.driver = {"name": disk_driver, "type": disk_format}
                    ret = virsh.attach_disk(vm.name, disk["source"], device_target[disk_num], attach_option, debug=True)
                    if not self.ignore_status:
                        libvirt.check_exit_status(ret, disk_attach_error)
                    if detach_option:
                        ret = virsh.detach_disk(vm.name, device_target[disk_num], extra=detach_option)
                        if not self.ignore_status:
                            libvirt.check_exit_status(ret)
                        libvirt.delete_local_disk(disk_type, disk_name)
            else:
                raise NotImplementedError
            time.sleep(self.itr_sleep_time)

    def host_stress_event(self, event):
        """
        Host Stress events

        :param event: event name
        """
        for itr in range(self.host_iterations):
            if "cpu_freq_governor" in event:
                cpu.set_freq_governor() if hasattr(cpu, 'set_freq_governor') else cpu.set_cpufreq_governor()
                logging.debug("Current governor: %s", cpu.get_freq_governor() if hasattr(cpu, 'get_freq_governor') else cpu.get_cpufreq_governor())
                time.sleep(self.event_sleep_time)
            elif "cpu_idle" in event:
                idlestate = cpu.get_idle_state() if hasattr(cpu, 'get_idle_state') else cpu.get_cpuidle_state()
                cpu.set_idle_state() if hasattr(cpu, 'set_idle_state') else cpu.set_cpuidle_state()
                time.sleep(self.event_sleep_time)
                cpu.set_idle_state(setstate=idlestate) if hasattr(cpu, 'set_idle_state') else cpu.set_cpuidle_state(setstate=idlestate)
                time.sleep(self.event_sleep_time)
            elif "cpuoffline" in event:
                online_count = cpu.online_count() if hasattr(cpu, 'online_count') else cpu.online_cpus_count()
                processor = self.host_cpu_list[random.randint(0, online_count-1)]
                cpu.offline(processor)
                time.sleep(self.event_sleep_time)
                cpu.online(processor)
            else:
                raise NotImplementedError
            time.sleep(self.itr_sleep_time)



from virttest import utils_misc


class Des(object):

    """
    Base Data Encryption Standard class.
    For details, please refer to:
    http://en.wikipedia.org/wiki/Data_Encryption_Standard
    """
    # Permutation and translation tables for DES
    PC1 = [
        56, 48, 40, 32, 24, 16, 8,
        0, 57, 49, 41, 33, 25, 17,
        9, 1, 58, 50, 42, 34, 26,
        18, 10, 2, 59, 51, 43, 35,
        62, 54, 46, 38, 30, 22, 14,
        6, 61, 53, 45, 37, 29, 21,
        13, 5, 60, 52, 44, 36, 28,
        20, 12, 4, 27, 19, 11, 3
    ]

    # Number left rotations of pc1
    left_rotations = [
        1, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1
    ]

    # get_sub_listd choice key (table 2)
    PC2 = [
        13, 16, 10, 23, 0, 4,
        2, 27, 14, 5, 20, 9,
        22, 18, 11, 3, 25, 7,
        15, 6, 26, 19, 12, 1,
        40, 51, 30, 36, 46, 54,
        29, 39, 50, 44, 32, 47,
        43, 48, 38, 55, 33, 52,
        45, 41, 49, 35, 28, 31
    ]

    # Initial permutation IP
    IP = [
        57, 49, 41, 33, 25, 17, 9, 1,
        59, 51, 43, 35, 27, 19, 11, 3,
        61, 53, 45, 37, 29, 21, 13, 5,
        63, 55, 47, 39, 31, 23, 15, 7,
        56, 48, 40, 32, 24, 16, 8, 0,
        58, 50, 42, 34, 26, 18, 10, 2,
        60, 52, 44, 36, 28, 20, 12, 4,
        62, 54, 46, 38, 30, 22, 14, 6
    ]

    # Expansion table for turning 32 bit blocks into 48 bits
    E = [
        31, 0, 1, 2, 3, 4,
        3, 4, 5, 6, 7, 8,
        7, 8, 9, 10, 11, 12,
        11, 12, 13, 14, 15, 16,
        15, 16, 17, 18, 19, 20,
        19, 20, 21, 22, 23, 24,
        23, 24, 25, 26, 27, 28,
        27, 28, 29, 30, 31, 0
    ]

    # The (in)famous S-boxes
    sbox = [
        # S1
        [14, 4, 13, 1, 2, 15, 11, 8, 3, 10, 6, 12, 5, 9, 0, 7,
         0, 15, 7, 4, 14, 2, 13, 1, 10, 6, 12, 11, 9, 5, 3, 8,
         4, 1, 14, 8, 13, 6, 2, 11, 15, 12, 9, 7, 3, 10, 5, 0,
         15, 12, 8, 2, 4, 9, 1, 7, 5, 11, 3, 14, 10, 0, 6, 13],

        # S2
        [15, 1, 8, 14, 6, 11, 3, 4, 9, 7, 2, 13, 12, 0, 5, 10,
         3, 13, 4, 7, 15, 2, 8, 14, 12, 0, 1, 10, 6, 9, 11, 5,
         0, 14, 7, 11, 10, 4, 13, 1, 5, 8, 12, 6, 9, 3, 2, 15,
         13, 8, 10, 1, 3, 15, 4, 2, 11, 6, 7, 12, 0, 5, 14, 9],

        # S3
        [10, 0, 9, 14, 6, 3, 15, 5, 1, 13, 12, 7, 11, 4, 2, 8,
         13, 7, 0, 9, 3, 4, 6, 10, 2, 8, 5, 14, 12, 11, 15, 1,
         13, 6, 4, 9, 8, 15, 3, 0, 11, 1, 2, 12, 5, 10, 14, 7,
         1, 10, 13, 0, 6, 9, 8, 7, 4, 15, 14, 3, 11, 5, 2, 12],

        # S4
        [7, 13, 14, 3, 0, 6, 9, 10, 1, 2, 8, 5, 11, 12, 4, 15,
         13, 8, 11, 5, 6, 15, 0, 3, 4, 7, 2, 12, 1, 10, 14, 9,
         10, 6, 9, 0, 12, 11, 7, 13, 15, 1, 3, 14, 5, 2, 8, 4,
         3, 15, 0, 6, 10, 1, 13, 8, 9, 4, 5, 11, 12, 7, 2, 14],

        # S5
        [2, 12, 4, 1, 7, 10, 11, 6, 8, 5, 3, 15, 13, 0, 14, 9,
         14, 11, 2, 12, 4, 7, 13, 1, 5, 0, 15, 10, 3, 9, 8, 6,
         4, 2, 1, 11, 10, 13, 7, 8, 15, 9, 12, 5, 6, 3, 0, 14,
         11, 8, 12, 7, 1, 14, 2, 13, 6, 15, 0, 9, 10, 4, 5, 3],

        # S6
        [12, 1, 10, 15, 9, 2, 6, 8, 0, 13, 3, 4, 14, 7, 5, 11,
         10, 15, 4, 2, 7, 12, 9, 5, 6, 1, 13, 14, 0, 11, 3, 8,
         9, 14, 15, 5, 2, 8, 12, 3, 7, 0, 4, 10, 1, 13, 11, 6,
         4, 3, 2, 12, 9, 5, 15, 10, 11, 14, 1, 7, 6, 0, 8, 13],

        # S7
        [4, 11, 2, 14, 15, 0, 8, 13, 3, 12, 9, 7, 5, 10, 6, 1,
         13, 0, 11, 7, 4, 9, 1, 10, 14, 3, 5, 12, 2, 15, 8, 6,
         1, 4, 11, 13, 12, 3, 7, 14, 10, 15, 6, 8, 0, 5, 9, 2,
         6, 11, 13, 8, 1, 4, 10, 7, 9, 5, 0, 15, 14, 2, 3, 12],

        # S8
        [13, 2, 8, 4, 6, 15, 11, 1, 10, 9, 3, 14, 5, 0, 12, 7,
         1, 15, 13, 8, 10, 3, 7, 4, 12, 5, 6, 11, 0, 14, 9, 2,
         7, 11, 4, 1, 9, 12, 14, 2, 0, 6, 10, 13, 15, 3, 5, 8,
         2, 1, 14, 7, 4, 10, 8, 13, 15, 12, 9, 0, 3, 5, 6, 11],
    ]

    # 32-bit permutation function P used on the output of the S-boxes
    P = [
        15, 6, 19, 20, 28, 11,
        27, 16, 0, 14, 22, 25,
        4, 17, 30, 9, 1, 7,
        23, 13, 31, 26, 2, 8,
        18, 12, 29, 5, 21, 10,
        3, 24
    ]

    # Final permutation IP^-1
    FP = [
        39, 7, 47, 15, 55, 23, 63, 31,
        38, 6, 46, 14, 54, 22, 62, 30,
        37, 5, 45, 13, 53, 21, 61, 29,
        36, 4, 44, 12, 52, 20, 60, 28,
        35, 3, 43, 11, 51, 19, 59, 27,
        34, 2, 42, 10, 50, 18, 58, 26,
        33, 1, 41, 9, 49, 17, 57, 25,
        32, 0, 40, 8, 48, 16, 56, 24
    ]

    # Initialisation
    def __init__(self, key):
        """
        Initialize the instance.

        :param key: Original used in DES.
        """
        if len(key) != 8:
            key = (key + '\0' * 8)[:8]

        self.L = []
        self.R = []
        self.Kn = [[0] * 48] * 16    # 16 48-bit keys (K1 - K16)

        self.setKey(key)

    def getKey(self):
        """
        Just get the crypting key.
        """
        return self.key

    def setKey(self, key):
        """
        Will set the crypting key for this object.
        RFB protocol for authentication requires client to encrypt
        challenge sent by server with password using DES method. However,
        bits in each byte of the password are put in reverse order before
        using it as encryption key.

        :param key: Original used in DES.
        """
        newkey = []
        for ki in range(len(key)):
            bsrc = ord(key[ki])
            btgt = 0
            for i in range(8):
                if bsrc & (1 << i):
                    btgt = btgt | (1 << 7 - i)

            newkey.append(chr(btgt))
        self.key = newkey
        self.create_Kn()

    def get_sub_list(self, table, block):
        """
        Return sub list of block according to index in table.

        :param table: Index list.
        :param block: bit list used to get sub list.
        """
        block_list = []
        for x in table:
            block_list.append(block[x])
        return block_list

    def create_Kn(self):
        """
        Create the 16 subkeys,from K[0] to K[15], from the given key
        """
        key = self.get_sub_list(
            self.PC1, utils_misc.string_to_bitlist(self.getKey()))
        self.L = key[:28]
        self.R = key[28:]
        for i in range(16):
            # Perform circular left shifts
            for j in range(self.left_rotations[i]):
                self.L.append(self.L[0])
                del self.L[0]
                self.R.append(self.R[0])
                del self.R[0]
            # Create one of the 16 subkeys through pc2 permutation
            self.Kn[i] = self.get_sub_list(self.PC2, self.L + self.R)

    def f(self, K):
        """
        The Feistel function (F-function) of DES, operates on half a block
        (32 bits) at a time and consists of four stages:
        1. Expansion
        2. Key mixing
        3. Substitution
        4. Permutation

        :param K: One of sixteen 48-bit subkeys are derived from the main key.
        """
        # Expansion:
        # The 32-bit half-block is expanded to 48 bits using E.
        self.R = self.get_sub_list(self.E, self.R)

        # Key mixing: The result is combined with a subkey using an XOR
        # operation. Sixteen 48-bit subkeys are derived from the main key.
        self.R = list(map(lambda x, y: x ^ y, self.R, K))

        # The block is divided into eight 6-bit pieces
        B = [self.R[:6], self.R[6:12], self.R[12:18], self.R[18:24],
             self.R[24:30], self.R[30:36], self.R[36:42], self.R[42:]]

        # Substitution:
        Bn = [0] * 32
        pos = 0
        for j in range(8):
            # Work out the offsets
            m = (B[j][0] << 1) + B[j][5]
            n = (B[j][1] << 3) + (B[j][2] << 2) + (B[j][3] << 1) + B[j][4]

            # Find the permutation value
            v = self.sbox[j][(m << 4) + n]

            # Turn value into bits, add it to result: Bn
            Bn[pos] = (v & 8) >> 3
            Bn[pos + 1] = (v & 4) >> 2
            Bn[pos + 2] = (v & 2) >> 1
            Bn[pos + 3] = v & 1

            pos += 4

        # Permutation:
        # Bn are rearranged according to a fixed permutation, the P-box.
        self.R = self.get_sub_list(self.P, Bn)

    def des_crypt(self, data, crypt_type=0):
        """
        Crypt the block of data through DES bit-manipulation

        :param data: data need to crypt.
        :param crypt_type: crypt type. 0 means encrypt, and 1 means decrypt.
        """
        # Get new block by using Ip.
        block = self.get_sub_list(self.IP, data)
        self.L = block[:32]
        self.R = block[32:]

        if crypt_type == 0:
            des_i = 0
            des_adj = 1
        else:
            des_i = 15
            des_adj = -1

        i = 0
        while i < 16:
            tempR = self.R
            self.f(self.Kn[des_i])

            # Xor with L[i - 1]
            self.R = list(map(lambda x, y: x ^ y, self.R, self.L))
            # Optimization: This now replaces the below commented code
            self.L = tempR

            i += 1
            des_i += des_adj

        # Final permutation of R[16]L[16]
        final = self.get_sub_list(self.FP, self.R + self.L)
        return final

    def crypt(self, data, crypt_type=0):
        """
        Crypt the data in blocks, running it through des_crypt()

        :param data: Data to be encrypted/decrypted.
        :param crypt_type: crypt type. 0 means encrypt, and 1 means decrypt.
        """

        # Split the data into list, crypting each one separately
        i = 0
        result = []
        while i < len(data):
            # Test code for caching encryption results
            block = utils_misc.string_to_bitlist(data[i:i + 8])
            pro_block = self.des_crypt(block, crypt_type)

            # Add the resulting block to our list
            result.append(utils_misc.bitlist_to_string(pro_block))
            i += 8

        # Return the full encrypted/decrypted string
        return ''.join(result)



"""
Address device / device descriptor class

http://libvirt.org/formatdomain.html#elementsAddress
"""

from virttest.libvirt_xml import accessors, xcepts
from virttest.libvirt_xml.devices import base


class Address(base.TypedDeviceBase):

    __slots__ = ('attrs',)

    def __init__(self, type_name, virsh_instance=base.base.virsh):
        # Blindly accept any/all attributes as simple dictionary
        accessors.XMLElementDict('attrs', self, parent_xpath='/',
                                 tag_name='address')
        super(self.__class__, self).__init__(device_tag='address',
                                             type_name=type_name,
                                             virsh_instance=virsh_instance)

    @classmethod
    def new_from_dict(cls, attributes, virsh_instance=base.base.virsh):
        # type_name is manditory, throw exception if doesn't exist
        try:
            # pop() so don't process again in loop below
            instance = cls(type_name=attributes.pop('type_name'),
                           virsh_instance=virsh_instance)
        except (KeyError, AttributeError):
            raise xcepts.LibvirtXMLError("type_name is manditory for "
                                         "Address class")
        # Stick property values in as attributes
        xtfroot = instance.xmltreefile.getroot()
        for key, value in list(attributes.items()):
            xtfroot.set(key, value)
        return instance

    @classmethod
    def new_from_element(cls, element, virsh_instance=base.base.virsh):
        # element uses type attribute, class uses type_name
        edict = dict(list(element.items()))
        try:
            edict['type_name'] = edict.pop('type')
        except (KeyError, AttributeError):
            raise xcepts.LibvirtXMLError("type attribute is manditory for "
                                         "Address class")
        return cls.new_from_dict(edict, virsh_instance=virsh_instance)



"""
Accommodate libvirt pci controller utility functions.

:Copyright: 2020 Red Hat Inc.
"""

import logging


def get_max_contr_indexes(vm_xml, cntlr_type, cntlr_model, cntl_num=1):
    """
    Obtain a number of controllers' max indexes in given guest xml

    :param vm_xml: guest xml
    :param cntlr_type: controller type, like pci
    :param cntlr_model: controller model, like pcie-root-port
    :param cntl_num: number of indexes will be returned
    :return: list of max index numbers
    """
    usable_indexes = []
    for elem in vm_xml.devices.by_device_tag('controller'):
        if (cntlr_type == elem.type and cntlr_model == elem.model):
            usable_indexes.append(elem.index)
    usable_indexes = sorted(usable_indexes, reverse=True)

    logging.debug("The indexes returned for controller type '{}' and "
                  "controller model '{}' is '{}'".format(cntlr_type,
                                                         cntlr_model,
                                                         usable_indexes[:cntl_num]))
    return usable_indexes[:cntl_num]



"""
video device support class(es)

http://libvirt.org/formatdomain.html#elementsVideo
"""

from virttest.libvirt_xml import accessors
from virttest.libvirt_xml.devices import base


class Video(base.UntypedDeviceBase):

    __slots__ = ('model_type', 'model_ram', 'model_vram', 'model_heads',
                 'model_vgamem', 'model_vram64',
                 'primary', 'acceleration', 'address')

    def __init__(self, virsh_instance=base.base.virsh):
        accessors.XMLAttribute('model_type', self,
                               parent_xpath='/',
                               tag_name='model',
                               attribute='type')
        accessors.XMLAttribute('model_ram', self,
                               parent_xpath='/',
                               tag_name='model',
                               attribute='ram')
        accessors.XMLAttribute('model_vram', self,
                               parent_xpath='/',
                               tag_name='model',
                               attribute='vram')
        accessors.XMLAttribute('model_vgamem', self,
                               parent_xpath='/',
                               tag_name='model',
                               attribute='vgamem')
        accessors.XMLAttribute('model_vram64', self,
                               parent_xpath='/',
                               tag_name='model',
                               attribute='vram64')
        accessors.XMLAttribute('model_heads', self,
                               parent_xpath='/',
                               tag_name='model',
                               attribute='heads')
        accessors.XMLAttribute('primary', self,
                               parent_xpath='/',
                               tag_name='model',
                               attribute='primary')
        accessors.XMLElementDict('acceleration', self,
                                 parent_xpath='/model',
                                 tag_name='acceleration')
        accessors.XMLElementDict('address', self,
                                 parent_xpath='/',
                                 tag_name='address')
        super(Video, self).__init__(device_tag='video',
                                    virsh_instance=virsh_instance)



"""
High-level QEMU test utility functions.

This module is meant to reduce code size by performing common test procedures.
Generally, code here should look like test code.

More specifically:
    - Functions in this module should raise exceptions if things go wrong
    - Functions in this module typically use functions and classes from
      lower-level modules (e.g. utils_misc, qemu_vm, aexpect).
    - Functions in this module should not be used by lower-level modules.
    - Functions in this module should be used in the right context.
      For example, a function should not be used where it may display
      misleading or inaccurate info or debug messages.

:copyright: 2008-2013 Red Hat Inc.
"""

import os
import re
import six
import time
import logging
from functools import reduce

from avocado.core import exceptions
from avocado.utils import path as utils_path
from avocado.utils import process
from avocado.utils import cpu as cpuutil

from virttest import error_context
from virttest import utils_misc
from virttest import qemu_monitor
from virttest.qemu_devices import qdevices
from virttest.staging import utils_memory


def guest_active(vm):
    o = vm.monitor.info("status")
    if isinstance(o, six.string_types):
        return "status: running" in o
    else:
        if "status" in o:
            return o.get("status") == "running"
        else:
            return o.get("running")


def get_numa_status(numa_node_info, qemu_pid, debug=True):
    """
    Get the qemu process memory use status and the cpu list in each node.

    :param numa_node_info: Host numa node information
    :type numa_node_info: NumaInfo object
    :param qemu_pid: process id of qemu
    :type numa_node_info: string
    :param debug: Print the debug info or not
    :type debug: bool
    :return: memory and cpu list in each node
    :rtype: tuple
    """
    node_list = numa_node_info.online_nodes
    qemu_memory = []
    qemu_cpu = []
    cpus = cpuutil.get_pid_cpus(qemu_pid)
    for node_id in node_list:
        qemu_memory_status = utils_memory.read_from_numa_maps(qemu_pid,
                                                              "N%d" % node_id)
        memory = sum([int(_) for _ in list(qemu_memory_status.values())])
        qemu_memory.append(memory)
        cpu = [_ for _ in cpus if _ in numa_node_info.nodes[node_id].cpus]
        qemu_cpu.append(cpu)
        if debug:
            logging.debug("qemu-kvm process using %s pages and cpu %s in "
                          "node %s" % (memory, " ".join(cpu), node_id))
    return (qemu_memory, qemu_cpu)


def pin_vm_threads(vm, node):
    """
    Pin VM threads to single cpu of a numa node

    :param vm: VM object
    :param node: NumaNode object
    """
    if len(vm.vcpu_threads) + len(vm.vhost_threads) < len(node.cpus):
        for i in vm.vcpu_threads:
            logging.info("pin vcpu thread(%s) to cpu(%s)" %
                         (i, node.pin_cpu(i)))
        for i in vm.vhost_threads:
            logging.info("pin vhost thread(%s) to cpu(%s)" %
                         (i, node.pin_cpu(i)))
    elif (len(vm.vcpu_threads) <= len(node.cpus) and
          len(vm.vhost_threads) <= len(node.cpus)):
        for i in vm.vcpu_threads:
            logging.info("pin vcpu thread(%s) to cpu(%s)" %
                         (i, node.pin_cpu(i)))
        for i in vm.vhost_threads:
            logging.info("pin vhost thread(%s) to extra cpu(%s)" %
                         (i, node.pin_cpu(i, extra=True)))
    else:
        logging.info("Skip pinning, no enough nodes")


def _check_driver_verifier(session, driver, verifier_flags=None, timeout=300):
    """
    Check driver verifier status

    :param session: VM session.
    :param driver: The driver need to query
    :param timeout: Timeout in seconds
    """
    logging.info("Check %s driver verifier status" % driver)
    query_cmd = "verifier /querysettings"
    output = session.cmd_output(query_cmd, timeout=timeout)
    status = driver in output
    if verifier_flags:
        status &= bool(re.findall(r"%s" % verifier_flags, output, re.I))
    return (status, output)


@error_context.context_aware
def setup_win_driver_verifier(session, driver, vm, timeout=300):
    """
    Enable driver verifier for windows guest.

    :param driver: The driver which needs enable the verifier.
    :param vm: VM object.
    :param timeout: Timeout in seconds.
    """

    win_verifier_flags = vm.params.get("windows_verifier_flags")
    verifier_status = _check_driver_verifier(session, driver,
                                             win_verifier_flags)[0]
    if not verifier_status:
        error_context.context("Enable %s driver verifier" % driver,
                              logging.info)
        if win_verifier_flags:
            verifier_setup_cmd = "verifier /flags %s /driver %s.sys" % (
                                 win_verifier_flags, driver)
        else:
            verifier_setup_cmd = "verifier /standard /driver %s.sys" % driver
        session.cmd(verifier_setup_cmd, timeout=timeout, ignore_all_errors=True)
        session = vm.reboot(session)
        verifier_status, output = _check_driver_verifier(session, driver,
                                                         win_verifier_flags)
        if not verifier_status:
            msg = "%s verifier is not enabled, details: %s" % (driver, output)
            raise exceptions.TestFail(msg)
    logging.info("%s verifier is enabled already" % driver)
    return session


def clear_win_driver_verifier(driver, vm, timeout=300):
    """
    Clear the driver verifier in windows guest.

    :param driver: The driver need to clear
    :param vm: VM object.
    :param timeout: Timeout in seconds.
    """
    session = vm.wait_for_login(timeout=timeout)
    try:
        verifier_status = _check_driver_verifier(session, driver)[1]
        if verifier_status:
            logging.info("Clear driver verifier")
            verifier_clear_cmd = "verifier /reset"
            session.cmd(verifier_clear_cmd,
                        timeout=timeout,
                        ignore_all_errors=True)
            session = vm.reboot(session)
    finally:
        session.close()


@error_context.context_aware
def windrv_verify_running(session, test, driver, timeout=300):
    """
    Check if driver is running for windows guest within a period time.

    :param session: VM session
    :param test: Kvm test object
    :param driver: The driver which needs to check.
    :param timeout: Timeout in seconds.
    """

    def _check_driver_stat():
        """
        Check if driver is in Running status.

        """
        output = session.cmd_output(driver_check_cmd, timeout=timeout)
        if "Running" in output:
            return True
        return False

    error_context.context("Check %s driver state." % driver, logging.info)
    driver_check_cmd = (r'wmic sysdriver where PathName="C:\\Windows\\System32'
                        r'\\drivers\\%s.sys" get State /value') % driver

    if not utils_misc.wait_for(_check_driver_stat, timeout, 0, 5):
        test.error("%s driver is not running" % driver)


@error_context.context_aware
def windrv_check_running_verifier(session, vm, test, driver, timeout=300):
    """
    Check whether the windows driver is running, then enable driver verifier.

    :param vm: the VM that use the driver.
    :param test: the KVM test object.
    :param driver: the driver concerned.
    :timeout: the timeout to use in this process, in seconds.
    """
    windrv_verify_running(session, test, driver, timeout)
    return setup_win_driver_verifier(session, driver, vm, timeout)


def setup_runlevel(params, session):
    """
    Setup the runlevel in guest.

    :param params: Dictionary with the test parameters.
    :param session: VM session.
    """
    cmd = "runlevel"
    ori_runlevel = "0"
    expect_runlevel = params.get("expect_runlevel", "3")

    # Note: All guest services may have not been started when
    #       the guest gets IP addr; the guest runlevel maybe
    #       is "unknown" whose exit status is 1 at that time,
    #       which will cause the cmd execution failed. Need some
    #       time here to wait for the guest services start.
    if utils_misc.wait_for(lambda: session.cmd_status(cmd) == 0, 15):
        ori_runlevel = session.cmd(cmd)

    ori_runlevel = ori_runlevel.split()[-1]
    if ori_runlevel == expect_runlevel:
        logging.info("Guest runlevel is already %s as expected" % ori_runlevel)
    else:
        session.cmd("init %s" % expect_runlevel)
        tmp_runlevel = session.cmd(cmd)
        tmp_runlevel = tmp_runlevel.split()[-1]
        if tmp_runlevel != expect_runlevel:
            logging.warn("Changing runlevel from %s to %s failed (%s)!",
                         ori_runlevel, expect_runlevel, tmp_runlevel)


class GuestSuspend(object):

    """
    Suspend guest, supports both Linux and Windows.

    """
    SUSPEND_TYPE_MEM = "mem"
    SUSPEND_TYPE_DISK = "disk"

    def __init__(self, test, params, vm):
        if not params or not vm:
            raise exceptions.TestError("Missing 'params' or 'vm' parameters")

        self._open_session_list = []
        self.test = test
        self.vm = vm
        self.params = params
        self.login_timeout = float(self.params.get("login_timeout", 360))
        self.services_up_timeout = float(self.params.get("services_up_timeout",
                                                         30))
        self.os_type = self.params.get("os_type")

    def _get_session(self):
        self.vm.verify_alive()
        session = self.vm.wait_for_login(timeout=self.login_timeout)
        return session

    def _session_cmd_close(self, session, cmd):
        try:
            return session.cmd_status_output(cmd)
        finally:
            try:
                session.close()
            except Exception:
                pass

    def _cleanup_open_session(self):
        try:
            for s in self._open_session_list:
                if s:
                    s.close()
        except Exception:
            pass

    @error_context.context_aware
    def setup_bg_program(self, **args):
        """
        Start up a program as a flag in guest.
        """
        suspend_bg_program_setup_cmd = args.get("suspend_bg_program_setup_cmd")

        error_context.context(
            "Run a background program as a flag", logging.info)
        session = self._get_session()
        self._open_session_list.append(session)

        logging.debug("Waiting all services in guest are fully started.")
        time.sleep(self.services_up_timeout)

        session.sendline(suspend_bg_program_setup_cmd)

    @error_context.context_aware
    def check_bg_program(self, **args):
        """
        Make sure the background program is running as expected
        """
        suspend_bg_program_chk_cmd = args.get("suspend_bg_program_chk_cmd")

        error_context.context(
            "Verify background program is running", logging.info)
        session = self._get_session()
        s, _ = self._session_cmd_close(session, suspend_bg_program_chk_cmd)
        if s:
            raise exceptions.TestFail(
                "Background program is dead. Suspend failed.")

    @error_context.context_aware
    def kill_bg_program(self, **args):
        error_context.context("Kill background program after resume")
        suspend_bg_program_kill_cmd = args.get("suspend_bg_program_kill_cmd")

        try:
            session = self._get_session()
            self._session_cmd_close(session, suspend_bg_program_kill_cmd)
        except Exception as e:
            logging.warn("Could not stop background program: '%s'", e)
            pass

    @error_context.context_aware
    def _check_guest_suspend_log(self, **args):
        error_context.context("Check whether guest supports suspend",
                              logging.info)
        suspend_support_chk_cmd = args.get("suspend_support_chk_cmd")

        session = self._get_session()
        s, o = self._session_cmd_close(session, suspend_support_chk_cmd)

        return s, o

    def verify_guest_support_suspend(self, **args):
        s, _ = self._check_guest_suspend_log(**args)
        if s:
            raise exceptions.TestError("Guest doesn't support suspend.")

    @error_context.context_aware
    def start_suspend(self, **args):
        suspend_start_cmd = args.get("suspend_start_cmd")
        error_context.context(
            "Start suspend [%s]" % (suspend_start_cmd), logging.info)

        session = self._get_session()
        self._open_session_list.append(session)

        # Suspend to disk
        session.sendline(suspend_start_cmd)

    @error_context.context_aware
    def verify_guest_down(self, **args):
        # Make sure the VM goes down
        error_context.context("Wait for guest goes down after suspend")
        suspend_timeout = 240 + int(self.params.get("smp")) * 60
        if not utils_misc.wait_for(self.vm.is_dead, suspend_timeout, 2, 2):
            raise exceptions.TestFail("VM refuses to go down. Suspend failed.")

    @error_context.context_aware
    def resume_guest_mem(self, **args):
        error_context.context("Resume suspended VM from memory")
        self.vm.monitor.system_wakeup()

    @error_context.context_aware
    def resume_guest_disk(self, **args):
        error_context.context("Resume suspended VM from disk")
        self.vm.create()

    @error_context.context_aware
    def verify_guest_up(self, **args):
        error_context.context("Verify guest system log", logging.info)
        suspend_log_chk_cmd = args.get("suspend_log_chk_cmd")

        session = self._get_session()
        s, o = self._session_cmd_close(session, suspend_log_chk_cmd)
        if s:
            raise exceptions.TestError(
                "Could not find suspend log. [%s]" % (o))

    @error_context.context_aware
    def action_before_suspend(self, **args):
        error_context.context("Actions before suspend")
        pass

    @error_context.context_aware
    def action_during_suspend(self, **args):
        error_context.context(
            "Sleep a while before resuming guest", logging.info)

        time.sleep(10)
        if self.os_type == "windows":
            # Due to WinXP/2003 won't suspend immediately after issue S3 cmd,
            # delay 10~60 secs here, maybe there's a bug in windows os.
            logging.info("WinXP/2003 need more time to suspend, sleep 50s.")
            time.sleep(50)

    @error_context.context_aware
    def action_after_suspend(self, **args):
        error_context.context("Actions after suspend")
        pass


class MemoryBaseTest(object):

    """
    Base class for memory functions.
    """

    UNIT = "M"

    def __init__(self, test, params, env):
        self.test = test
        self.env = env
        self.params = params
        self.sessions = {}

    def get_vm_mem(self, vm):
        """
        Count memory assigned to VM.

        :param vm: VM object
        :return: memory size in MB.
        """
        PC_DIMM = qdevices.Dimm
        # Default memory size in configuration is MB, so append
        # 'MB' in the end of 'mem' param.
        mem_str = "%sM" % vm.params.get("mem", "0")
        total_mem = self.normalize_mem_size(mem_str)
        pc_dimms = filter(lambda x: isinstance(x, PC_DIMM), vm.devices)
        obj_ids = map(lambda x: x.get_param('memdev'), pc_dimms)
        obj_devs = map(lambda x: vm.devices.get_by_qid(x)[0], obj_ids)
        obj_size = map(lambda x: x.get_param('size'), obj_devs)
        total_mem += sum(map(self.normalize_mem_size, obj_size))
        logging.info("Assigned %s%s " % (total_mem, self.UNIT) +
                     "memory to '%s'" % vm.name)
        return total_mem

    @classmethod
    def normalize_mem_size(cls, str_size):
        """
        Convert memory size unit

        :param str_size: memory size string, like: 1GB
        :return: memory size value in MB
        """
        args = (str_size, cls.UNIT, 1024)
        try:
            size = utils_misc.normalize_data_size(*args)
            return int(float(size))
        except ValueError as details:
            logging.debug("Convert memory size error('%s')" % details)
        return 0

    @classmethod
    def get_guest_total_mem(cls, vm):
        """
        Guest OS reported physical memory size in MB.

        :param vm: VM object.
        :return: physical memory report by guest OS in MB
        """
        if vm.params.get("os_type") == "windows":
            cmd = 'wmic ComputerSystem get TotalPhysicalMemory'
        else:
            cmd = "grep 'MemTotal:' /proc/meminfo"
        return vm.get_memory_size(cmd)

    @classmethod
    def get_guest_free_mem(cls, vm):
        """
        Guest OS reported free memory size in MB.

        :param vm: VM Object
        :return: free memory report by guest OS in MB
        """
        os_type = vm.params.get("os_type")
        timeout = float(vm.params.get("login_timeout", 600))
        try:
            session = vm.wait_for_login(timeout=timeout)
            return utils_misc.get_free_mem(session, os_type)
        finally:
            session.close()

    @classmethod
    def get_guest_used_mem(cls, vm):
        """
        Guest OS reported used memory size in MB.

        :param vm: VM Object
        :return: used memory report by guest OS in MB
        """
        os_type = vm.params.get("os_type")
        timeout = float(vm.params.get("login_timeout", 600))
        session = vm.wait_for_login(timeout=timeout)
        try:
            return utils_misc.get_used_mem(session, os_type)
        finally:
            session.close()

    def get_session(self, vm):
        """
        Get connection to VM.

        :param vm: VM object
        :return: return ShellSession object
        """
        key = vm.instance
        self.sessions.setdefault(key, [])
        for session in self.sessions.get(key):
            if session.is_responsive():
                return session
            session.close()
            self.sessions[key].remove(session)
        login_timeout = float(self.params.get("login_timeout", 600))
        session = vm.wait_for_login(timeout=login_timeout)
        self.sessions[key].append(session)
        return session

    def close_sessions(self):
        """
        Close opening session, better to call it in the end of test.
        """
        sessions = list(filter(None, list(self.sessions.values())))
        if sessions:
            sessions = list(filter(None, reduce(list.__add__, sessions)))
            list(map(lambda x: x.close(), sessions))
        self.sessions.clear()


class MemoryHotplugTest(MemoryBaseTest):

    """
    Class for memory hotplug/unplug test.
    """

    @error_context.context_aware
    def update_vm_after_hotplug(self, vm, dev):
        """
        Update VM params to ensure hotpluged devices exist in guest.

        :param vm: VM object
        :param dev: Qdevice object.
        """
        error_context.context("Update VM object after hotplug memory")
        dev_type, name = dev.get_qid().split('-')
        if isinstance(dev, qdevices.Memory):
            backend = self.params.object_params(name).get("backend_mem",
                                                          "memory-backend-ram")
            attrs = dev.__attributes__[backend][:]
        else:
            attrs = dev.__attributes__[:]
        params = self.params.copy_from_keys(attrs)
        for attr in attrs:
            val = dev.get_param(attr)
            if val:
                key = "_".join([attr, dev_type, name])
                params[key] = val
        mem_devs = vm.params.objects("mem_devs")
        if name not in mem_devs:
            mem_devs.append(name)
            params["mem_devs"] = " ".join(mem_devs)
        vm.params.update(params)
        if dev not in vm.devices:
            vm.devices.insert(dev)
        self.env.register_vm(vm.name, vm)

    @error_context.context_aware
    def update_vm_after_unplug(self, vm, dev):
        """
        Update VM params object after unplug memory devices.

        :param vm: VM object
        :param dev: Qdevice object
        """
        error_context.context("Update VM object after unplug memory")
        dev_type, name = dev.get_qid().split('-')
        mem_devs = vm.params.objects("mem_devs")
        if dev in vm.devices:
            vm.devices.remove(dev)
        if name in mem_devs:
            mem_devs.remove(name)
            vm.params["mem_devs"] = " ".join(mem_devs)
        self.env.register_vm(vm.name, vm)

    @error_context.context_aware
    def hotplug_memory(self, vm, name):
        """
        Hotplug dimm device with memory backend

        :param vm: VM object
        :param name: memory device name
        """
        devices = vm.devices.memory_define_by_params(self.params, name)
        for dev in devices:
            dev_type = "memory"
            if isinstance(dev, qdevices.Dimm):
                addr = self.get_mem_addr(vm, dev.get_qid())
                dev.set_param("addr", addr)
                dev_type = "pc-dimm"
            step = "Hotplug %s '%s' to VM" % (dev_type, dev.get_qid())
            error_context.context(step, logging.info)
            _, ver_out = vm.devices.simple_hotplug(dev, vm.monitor)
            if ver_out is False:
                raise exceptions.TestFail("Verify hotplug memory failed")
            self.update_vm_after_hotplug(vm, dev)
        return devices

    @error_context.context_aware
    def unplug_memory(self, vm, name):
        """
        Unplug memory device
        step 1, unplug dimm device
        step 2, unplug memory object

        :param vm: VM object
        :param name: memory device name
        """
        devices = []
        qid_dimm = "dimm-%s" % name
        qid_mem = "mem-%s" % name
        try:
            dimm = vm.devices.get_by_qid(qid_dimm)[0]
        except IndexError:
            logging.warn("'%s' is not used by any dimm" % qid_mem)
        else:
            step = "Unplug pc-dimm '%s'" % qid_dimm
            error_context.context(step, logging.info)
            _, ver_out = vm.devices.simple_unplug(dimm, vm.monitor)
            if ver_out is False:
                raise exceptions.TestFail("Verify unplug memory failed")
            devices.append(dimm)
            self.update_vm_after_unplug(vm, dimm)

        step = "Unplug memory object '%s'" % qid_mem
        error_context.context(step, logging.info)
        try:
            mem = vm.devices.get_by_qid(qid_mem)[0]
        except IndexError:
            output = vm.monitor.query("memory-devices")
            logging.debug("Memory devices: %s" % output)
            msg = "Memory object '%s' not exists" % qid_mem
            raise exceptions.TestError(msg)
        error_context.context(step, logging.info)
        vm.devices.simple_unplug(mem, vm.monitor)
        devices.append(mem)
        self.update_vm_after_unplug(vm, mem)
        return devices

    @error_context.context_aware
    def get_mem_addr(self, vm, qid):
        """
        Get guest memory address from qemu monitor

        :param vm: VM object
        :param qid: memory device qid
        """
        error_context.context("Get hotpluged memory address", logging.info)
        if not isinstance(vm.monitor, qemu_monitor.QMPMonitor):
            raise NotImplementedError
        for info in vm.monitor.info("memory-devices"):
            if str(info['data']['id']) == qid:
                address = info['data']['addr']
                logging.info("Memory address: %s" % address)
                return address

    @error_context.context_aware
    def check_memory(self, vm=None):
        """
        Check is guest memory is really match assgined to VM.

        :param vm: VM object, get VM object from env if vm is None.
        """
        error_context.context("Verify memory info", logging.info)
        if not vm:
            vm = self.env.get_vm(self.params["main_vm"])
        vm.verify_alive()
        threshold = float(self.params.get("threshold", 0.10))
        timeout = float(self.params.get("wait_resume_timeout", 60))
        # Notes:
        #    some sub test will pause VM, here need to wait VM resume
        # then check memory info in guest.
        utils_misc.wait_for(lambda: not vm.is_paused(), timeout=timeout)
        utils_misc.verify_dmesg()
        self.os_type = self.params.get("os_type")
        guest_mem_size = super(MemoryHotplugTest, self).get_guest_total_mem(vm)
        vm_mem_size = self.get_vm_mem(vm)
        if abs(guest_mem_size - vm_mem_size) > vm_mem_size * threshold:
            msg = ("Assigned '%s MB' memory to '%s'"
                   "but, '%s MB' memory detect by OS" %
                   (vm_mem_size, vm.name, guest_mem_size))
            raise exceptions.TestFail(msg)

    @error_context.context_aware
    def memory_operate(self, vm, memory, operation='online'):
        error_context.context(
            "%s %s in guest OS" %
            (operation, memory), logging.info)
        mem_sys_path = "/sys/devices/system/memory/%s" % memory
        mem_state_path = os.path.join(mem_sys_path, 'state')
        session = self.get_session(vm)
        session.cmd("echo '%s' > %s" % (operation, mem_state_path))
        output = session.cmd_output_safe("cat %s" % mem_state_path)
        if operation not in output:
            return exceptions.TestFail("Fail to %s %s" % (operation, memory))

    def get_memory_state(self, vm, memory):
        """Get memorys state in guest OS"""
        mem_sys_path = "/sys/devices/system/memory/%s" % memory
        mem_state_path = os.path.join(mem_sys_path, 'state')
        session = self.get_session(vm)
        status, output = session.cmd_status_output("cat %s" % mem_state_path)
        if status != 0:
            raise exceptions.TestError("Fail to read %s state" % memory)
        return output.strip()

    def get_offline_memorys(self, vm):
        """Get unusable memory in guest OS"""
        def is_offline_memory(x):
            return self.get_memory_state(vm, x) == 'offline'

        memorys = self.get_all_memorys(vm)
        return set(filter(is_offline_memory, memorys))

    def get_all_memorys(self, vm):
        """Get all memorys detected in guest OS"""
        mem_sys_path = "/sys/devices/system/memory"
        cmd = "ls %s | grep memory" % mem_sys_path
        session = self.get_session(vm)
        output = session.cmd_output_safe(cmd, timeout=90)
        return set([_ for _ in output.splitlines() if _])



"""
Generic character device support for serial, parallel, channel, and console

http://libvirt.org/formatdomain.html#elementCharSerial
"""

from virttest.libvirt_xml.devices import base


class CharacterBase(base.TypedDeviceBase):

    __slots__ = ('sources', 'targets')

    # Not overriding __init__ because ABC cannot hide device_tag as expected

    # Accessors just wrap private helpers in UntypedDeviceBase class
    def get_sources(self):
        """
        Return a list of dictionaries containing each source's attributes.
        """
        return self._get_list('source')

    def set_sources(self, value):
        """
        Set all sources to the value list of dictionaries of source attributes.
        """
        self._set_list('source', value)

    def del_sources(self):
        """
        Remove the list of dictionaries containing each source's attributes.
        """
        self._del_list('source')

    def get_targets(self):
        """
        Return a list of dictionaries containing each target's attributes.
        """
        return self._get_list('target')

    def set_targets(self, value):
        """
        Set all sources to the value list of dictionaries of target attributes.
        """
        self._set_list('target', value)

    def del_targets(self):
        """
        Remove the list of dictionaries containing each target's attributes.
        """
        self._del_list('target')

    # Some convenience methods so appending to sources/targets is easier
    def add_source(self, **attributes):
        """
        Convenience method for appending a source from dictionary of attributes
        """
        self._add_item('sources', **attributes)

    def add_target(self, **attributes):
        """
        Convenience method for appending a target from dictionary of attributes
        """
        self._add_item('targets', **attributes)

    def update_source(self, index, **attributes):
        """
        Convenience method for merging values into a source's attributes
        """
        self._update_item('sources', index, **attributes)

    def update_target(self, index, **attributes):
        """
        Convenience method for merging values into a target's attributes
        """
        self._update_item('targets', index, **attributes)



"""
random number generator device support class(es)

http://libvirt.org/formatdomain.html#elementsRng
"""

from virttest.libvirt_xml import accessors, xcepts
from virttest.libvirt_xml.devices import base, librarian


class Rng(base.UntypedDeviceBase):

    __slots__ = ('rng_model', 'rate', 'backend', 'alias', 'address')

    def __init__(self, virsh_instance=base.base.virsh):
        accessors.XMLAttribute('rng_model', self,
                               parent_xpath='/',
                               tag_name='rng',
                               attribute='model')
        accessors.XMLElementDict('rate', self,
                                 parent_xpath='/',
                                 tag_name='rate')
        accessors.XMLElementNest('backend', self, parent_xpath='/',
                                 tag_name='backend', subclass=self.Backend,
                                 subclass_dargs={
                                     'virsh_instance': virsh_instance})
        accessors.XMLElementNest('address', self, parent_xpath='/',
                                 tag_name='address', subclass=self.Address,
                                 subclass_dargs={'type_name': 'pci',
                                                 'virsh_instance': virsh_instance})
        accessors.XMLElementDict('alias', self, parent_xpath='/',
                                 tag_name='alias')
        super(Rng, self).__init__(
            device_tag='rng', virsh_instance=virsh_instance)
        self.xml = '<rng/>'

    Address = librarian.get('address')

    def new_rng_address(self, **dargs):
        """
        Return a new interface Address instance and set properties from dargs
        """
        new_one = self.Address("pci", virsh_instance=self.virsh)
        for key, value in list(dargs.items()):
            setattr(new_one, key, value)
        return new_one

    class Backend(base.base.LibvirtXMLBase):

        """
        Rng backend xml class.

        Properties:

        model:
            string. backend model
        type:
            string. backend type
        """
        __slots__ = ('backend_model', 'backend_type', 'backend_dev',
                     'source', 'backend_protocol')

        def __init__(self, virsh_instance=base.base.virsh):
            accessors.XMLAttribute(property_name="backend_model",
                                   libvirtxml=self,
                                   forbidden=None,
                                   parent_xpath='/',
                                   tag_name='backend',
                                   attribute='model')
            accessors.XMLAttribute(property_name="backend_type",
                                   libvirtxml=self,
                                   forbidden=None,
                                   parent_xpath='/',
                                   tag_name='backend',
                                   attribute='type')
            accessors.XMLElementText('backend_dev',
                                     self, parent_xpath='/',
                                     tag_name='backend')

            accessors.XMLElementList(property_name='source',
                                     libvirtxml=self,
                                     parent_xpath='/',
                                     marshal_from=self.marshal_from_source,
                                     marshal_to=self.marshal_to_source)
            accessors.XMLAttribute(property_name="backend_protocol",
                                   libvirtxml=self,
                                   forbidden=None,
                                   parent_xpath='/',
                                   tag_name='protocol',
                                   attribute='type')
            super(self.__class__, self).__init__(virsh_instance=virsh_instance)
            self.xml = '<backend/>'

        @staticmethod
        def marshal_from_source(item, index, libvirtxml):
            """Convert a dictionary into a tag + attributes"""
            del index           # not used
            del libvirtxml      # not used
            if not isinstance(item, dict):
                raise xcepts.LibvirtXMLError("Expected a dictionary of host "
                                             "attributes, not a %s"
                                             % str(item))
            return ('source', dict(item))  # return copy of dict, not reference

        @staticmethod
        def marshal_to_source(tag, attr_dict, index, libvirtxml):
            """Convert a tag + attributes into a dictionary"""
            del index                    # not used
            del libvirtxml               # not used
            if tag != 'source':
                return None              # skip this one
            return dict(attr_dict)       # return copy of dict, not reference



"""
Module simplifying manipulation of XML described at
http://libvirt.org/formatsnapshot.html
"""

from virttest import xml_utils
from virttest.libvirt_xml import base, accessors
from virttest.libvirt_xml.devices.disk import Disk
from virttest.libvirt_xml import xcepts


class SnapshotXMLBase(base.LibvirtXMLBase):

    """
    Accessor methods for SnapshotXML class.

    Properties:
        snap_name:
            string, operates on snapshot name tag
        description:
            string, operates on snapshot description tag
        mem_snap_type:
            string, operates snapshot type under memory tag, 'internal',
            'external' or 'no'
        mem_file:
            string, operates snapshot file path under memory tag
        creation_time:
            string, operates on creationTime tag
        state:
            string, operates snapshot state tag
        parent_name:
            string, parent snapshot name tag under parent tag
    """

    __slots__ = ('snap_name', 'description', 'mem_snap_type', 'mem_file',
                 'creation_time', 'state', 'parent_name')

    def __init__(self, virsh_instance=base.virsh):
        accessors.XMLElementText('snap_name', self, parent_xpath='/',
                                 tag_name='name')
        accessors.XMLElementText('description', self, parent_xpath='/',
                                 tag_name='description')
        accessors.XMLAttribute('mem_snap_type', self, parent_xpath='/',
                               tag_name='memory', attribute='snapshot')
        accessors.XMLAttribute('mem_file', self, parent_xpath='/',
                               tag_name='memory', attribute='file')
        accessors.XMLElementText('creation_time', self, parent_xpath='/',
                                 tag_name='creationTime')
        accessors.XMLElementText('state', self, parent_xpath='/',
                                 tag_name='state')
        accessors.XMLElementText('parent_name', self, parent_xpath='/parent',
                                 tag_name='name')
        super(SnapshotXMLBase, self).__init__(virsh_instance=virsh_instance)


class SnapshotXML(SnapshotXMLBase):

    """
    Manipulators of a snapshot through it's XML definition.
    """

    __slots__ = []

    def __init__(self, virsh_instance=base.virsh):
        """
        Initialize new instance with empty XML
        """
        super(SnapshotXML, self).__init__(virsh_instance=virsh_instance)
        self.xml = u"<domainsnapshot><disks>\
                     </disks></domainsnapshot>"

    @staticmethod
    def new_from_snapshot_dumpxml(name, snap_name, virsh_instance=base.virsh):
        """
        Return new SnapshotXML instance from virsh snapshot-dumpxml command

        :param name: vm's name
        :param snap_name: snapshot name
        :param uuid: snapshot's uuid
        :param virsh_instance: virsh module or instance to use
        :return: New initialized SnapshotXML instance
        """
        snapshot_xml = SnapshotXML(virsh_instance=virsh_instance)
        result = virsh_instance.snapshot_dumpxml(name, snap_name)
        snapshot_xml['xml'] = result.stdout_text.strip()

        return snapshot_xml

    def set_disks(self, value_list):
        """
        Define disks based on contents of SnapDiskXML instance list

        :param value_list: SnapDiskXML instance list
        """
        for value in value_list:
            value_type = type(value)
            if not isinstance(value, self.SnapDiskXML):
                raise xcepts.LibvirtXMLError("Value %s Must be a instance of "
                                             "SnapDiskXML, not a %s"
                                             % (str(value), str(value_type)))
        # Start with clean slate
        exist_dev = self.xmltreefile.find('disks')
        if exist_dev is not None:
            self.del_disks()
        if len(value_list) > 0:
            disks_element = xml_utils.ElementTree.SubElement(
                self.xmltreefile.getroot(), 'disks')
            for disk in value_list:
                # Separate the element from the tree
                disk_element = disk.xmltreefile.getroot()
                disks_element.append(disk_element)
        self.xmltreefile.write()

    def del_disks(self):
        """
        Remove all disks
        """
        self.xmltreefile.remove_by_xpath('/disks', remove_all=True)
        self.xmltreefile.write()

    class SnapDiskXML(Disk):

        """
        Manipulators disk xml in snapshot xml definition.
        Most properties are inherit from parent class Disk.

        Properties:
            disk_name:
                string, operates on disk name under disk tag
        """

        __slots__ = Disk.__slots__ + ('disk_name', )

        def __init__(self, virsh_instance=base.virsh):
            """
            Initialize new instance with empty XML
            """
            accessors.XMLAttribute('disk_name', self, parent_xpath='/',
                                   tag_name='disk', attribute='name')
            super(self.__class__, self).__init__(virsh_instance=virsh_instance)
            self.xml = u"<disk></disk>"



"""
Higher order classes and functions for Libvirt Sandbox (lxc) container testing

:copyright: 2013 Red Hat Inc.
"""

import datetime
import time
import logging

from virttest import lvsb_base


# This utility function lets test-modules quickly create a list of all
# sandbox aggregate types, themselves containing a list of individual
# sandboxes.


def make_sandboxes(params, env, extra_ns=None):
    """
    Return list of instantiated lvsb_testsandboxes classes from params

    :param params: an undiluted Params instance
    :param env: the current env instance
    :param extra_ns: An extra, optional namespace to search for classes
    """
    namespace = globals()  # stuff in this module
    # For specialized sandbox types, allow their class to be defined
    # inside test module or elsewhere.
    if extra_ns is not None:
        namespace.update(extra_ns)  # copy in additional symbols
    names = list(namespace.keys())
    # Test may require more than one sandbox agregator class
    pobs = params.objects('lvsb_testsandboxes')  # manditory parameter
    # filter out non-TestSandboxes subclasses
    for name in names:
        try:
            if not issubclass(namespace[name], lvsb_base.TestSandboxes):
                # Working on name list, okay to modify dict
                del namespace[name]
        except TypeError:
            # Symbol wasn't a class, just ignore it
            pass
    # Return a list of instantiated sandbox_testsandboxes's classes
    return [namespace[type_name](params, env) for type_name in pobs]


# TestSandboxes subclasses defined below, or inside other namespaces like
# a test module.  They simply help the test-module iterate over many
# aggregate manager classes and the sandboxes they contain.

class TestBaseSandboxes(lvsb_base.TestSandboxes):

    """
    Simplistic sandbox aggregate manager
    """

    def __init__(self, params, env):
        """
        Initialize to run, all SandboxCommandBase's
        """
        super(TestBaseSandboxes, self).__init__(params, env)
        self.init_sandboxes()  # create instances of SandboxCommandBase
        # Point all of them at the same local uri
        self.for_each(lambda sb: sb.add_optarg('-c', self.uri))
        # The flag doesn't require sandbox name
        if not self.flag:
            # Use each instances name() method to produce name argument
            self.for_each(lambda sb: sb.add_optarg('-n', sb.name))

    def command_suffixes(self):
        """
        Append command after a --
        """
        # Command should follow after a --
        self.for_each(lambda sb: sb.add_mm())
        # Each one gets the same command (that's why it's simple)
        self.for_each(lambda sb: sb.add_pos(self.command))

    def results(self, each_timeout=5):
        """
        Run sandboxe(s), allowing each_timeout to complete, return output list
        """
        # Sandboxes run asynchronously, prevent them from running forever
        start = datetime.datetime.now()
        total_timeout_seconds = each_timeout * self.count
        timeout_at = start + datetime.timedelta(seconds=total_timeout_seconds)
        # No need to write a method just to call the run method
        self.for_each(lambda sb: sb.run())
        while datetime.datetime.now() < timeout_at:
            # Wait until number of running sandboxes is zero
            if bool(self.are_running()):
                time.sleep(0.1)  # Don't busy-wait
                continue
            else:  # none are running
                break
        # Needed for accurate time in logging message below
        end = datetime.datetime.now()
        # Needed for logging message if none exited before timeout
        still_running = self.are_running()
        # Cause all exited sessions to clean up when sb.stop() called
        self.for_each(lambda sb: sb.auto_clean(True))
        # If raise, auto_clean will make sure cleanup happens
        if bool(still_running):
            raise lvsb_base.SandboxException("%d of %d sandboxes are still "
                                             "running after "
                                             "the timeout of %d seconds."
                                             % (still_running,
                                                self.count,
                                                total_timeout_seconds))
        # Kill off all sandboxes, just to be safe
        self.for_each(lambda sb: sb.stop())
        logging.info("%d sandboxe(s) finished in %s", self.count,
                     end - start)
        # Return a list of stdout contents from each
        return self.for_each(lambda sb: sb.recv())


# TestBaseSandboxes subclasses which just runs simple default
# options with the same command.

class TestSimpleSandboxes(TestBaseSandboxes):

    """
    Executes a command with simple options
    """

    def __init__(self, params, env):
        """
        Initialize to run, all SandboxCommandBase's
        """
        super(TestSimpleSandboxes, self).__init__(params, env)
        # Appends command after options
        self.command_suffixes()


# TestBaseSandboxes subclasses which runs complex options and allows
# iterating for the options with the same command.

class TestComplexSandboxes(TestBaseSandboxes):

    """
    Executes a command with complex options
    """

    def __init__(self, params, env):
        super(TestComplexSandboxes, self).__init__(params, env)
        # Appends command options
        if self.opts:
            for k, v in self.opts:
                self.for_each(lambda sb: sb.add_optarg(k, v))
            # Appends command after options
            self.command_suffixes()
        if self.flag:
            for k in self.flag:
                self.for_each(lambda sb: sb.add_flag(k))
            # only '-h' and '-V' flags don't require '--' with command
            if "-h" not in self.flag and "-V" not in self.flag:
                # Appends command after options
                self.command_suffixes()



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2018
# Author: Lukas Doktor <ldoktor@redhat.com>

"""
Avocado plugin that extends the settings path of our config paths
"""

import os
from pkg_resources import resource_filename
from pkg_resources import resource_listdir

from avocado.core.plugin_interfaces import Settings


class VTSettings(Settings):

    def adjust_settings_paths(self, paths):
        base = resource_filename('avocado_vt', 'conf.d')
        for path in [os.path.join(base, conf)
                     for conf in resource_listdir('avocado_vt', 'conf.d')
                     if conf.endswith('.conf')]:
            paths.insert(0, path)



import os
import string

from virttest import error_context
from virttest import utils_misc

from avocado.utils import process


@error_context.context_aware
def get_image_filename(params):
    """
    Get the nbd image uri:
      tcp: nbd://<server-ip>[:<port>]/[<export>]
      unix domain socket: nbd+unix:///[<export>]?socket=<domain-socket>
    :param params: image specified params
    :return: nbd image uri string
    """
    server = params.get('nbd_server')
    unix_socket = params.get('nbd_unix_socket')
    export_name = '/%s' % params['nbd_export_name'] if params.get(
        'nbd_export_name') else ''

    if server:
        port = ':%s' % params['nbd_port'] if params.get('nbd_port') else ''
        return 'nbd://{server}{port}{export}'.format(server=server,
                                                     port=port,
                                                     export=export_name)
    elif unix_socket:
        return 'nbd+unix://{export}?socket={sock}'.format(export=export_name,
                                                          sock=unix_socket)

    raise ValueError('Either nbd_server or nbd_unix_socket is required.')


@error_context.context_aware
def export_image(qemu_nbd, filename, local_image, params):
    """
    Export a local file image with qemu-nbd command
    :param qemu_nbd: path to the qemu-nbd binary
    :param filename: image path for raw/qcow2 image or image json
                     representation string for luks image
    :param local_image: image tag defined in parameter images
    :param params: local image specified params
    :return: pid of qemu-nbd server or None
    """
    cmd_dict = {
        "secret_object": "",
        "tls_creds": "",
        "export_format": "",
        "persistent": "-t",
        "desc": "",
        "port": "",
        "export_name": "",
        "unix_socket": "",
        "filename": "",
        "fork": "--fork",
        "pid_file": "",
        "bitmap": ""
    }
    export_cmd = ('{secret_object} {tls_creds} '
                  '{export_format} {persistent} {desc} {port} {bitmap} '
                  '{export_name} {fork} {pid_file} {unix_socket} {filename}')

    pid_file = utils_misc.generate_tmp_file_name('%s_nbd_server' % local_image,
                                                 'pid')
    cmd_dict['pid_file'] = '--pid-file %s' % pid_file
    cmd_dict['filename'] = filename

    # auto-detect format if format(-f) is not specified
    if params.get('nbd_export_format'):
        cmd_dict['export_format'] = '-f %s' % params['nbd_export_format']

        if params['nbd_export_format'] == 'luks':
            # We can only export a local luks image in luks
            if params.get("image_format") != 'luks':
                raise ValueError("Only a luks image can be exported in luks")

            secret_str = '--object secret,id={aid},data={data}'
            cmd_dict.update({
                'secret_object': secret_str.format(
                    aid='%s_encrypt0' % local_image,
                    data=params["image_secret"]),
                'filename': "'%s'" % filename
            })

    if params.get('nbd_export_name'):
        cmd_dict['export_name'] = '-x %s' % params['nbd_export_name']

    if params.get('nbd_export_description'):
        cmd_dict['desc'] = '-D "%s"' % params['nbd_export_description']

    if params.get('nbd_unix_socket'):
        cmd_dict['unix_socket'] = '-k %s' % params['nbd_unix_socket']
    else:
        # 10809 is used by defalut if port is not set
        if params.get('nbd_port'):
            cmd_dict['port'] = '-p %s' % params['nbd_port']

        # tls creds is supported for ip only
        if params.get('nbd_server_tls_creds'):
            tls_str = ('--object tls-creds-x509,id={aid},endpoint=server,'
                       'dir={tls_creds} --tls-creds {aid}')
            cmd_dict['tls_creds'] = tls_str.format(
                aid='%s_server_tls_creds' % local_image,
                tls_creds=params['nbd_server_tls_creds']
            )

    if params.get('nbd_export_bitmap'):
        cmd_dict['bitmap'] = '-B %s' % params['nbd_export_bitmap']

    qemu_nbd_pid = None
    cmdline = qemu_nbd + ' ' + string.Formatter().format(export_cmd,
                                                         **cmd_dict)
    result = process.run(cmdline, ignore_status=True, shell=True,
                         ignore_bg_processes=True)
    if result.exit_status == 0:
        with open(pid_file, "r") as pid_file_fd:
            qemu_nbd_pid = int(pid_file_fd.read().strip())
        os.unlink(pid_file)

    return qemu_nbd_pid


def list_exported_image(qemu_nbd, nbd_image, params):
    """
    List all details about the exports by qemu-nbd
    :param qemu_nbd: path to the qemu-nbd binary
    :param nbd_image: nbd image tag defined in parameter images
    :param params: nbd image specified params
    :return: CmdResult object of qemu-nbd output
    """
    cmd_dict = {
        "tls_creds": "",
        "port": "",
        "server": ""
    }
    list_cmd = '-L {tls_creds} {port} {server}'

    if params.get('nbd_unix_socket'):
        cmd_dict['server'] = '-k %s' % params['nbd_unix_socket']
    else:
        # 0.0.0.0 is used by default if interface is not set
        if params.get('nbd_server'):
            cmd_dict['server'] = '-b %s' % params['nbd_server']

        # 10809 is used by default if port is not set
        if params.get('nbd_port'):
            cmd_dict['port'] = '-p %s' % params['nbd_port']

        # tls creds is supported for ip only
        if params.get('nbd_client_tls_creds'):
            tls_str = ('--object tls-creds-x509,id={aid},endpoint=client,'
                       'dir={tls_creds} --tls-creds {aid}')
            cmd_dict['tls_creds'] = tls_str.format(
                aid='%s_client_tls_creds' % nbd_image,
                tls_creds=params['nbd_client_tls_creds']
            )

    cmdline = qemu_nbd + ' ' + string.Formatter().format(list_cmd,
                                                         **cmd_dict)
    return process.run(cmdline, ignore_status=True, shell=True,
                       ignore_bg_processes=True)



import os
import re
import logging

from avocado.utils import process


class FS(object):
    """
    Base class for proc/sys FS set and get
    """
    def __init__(self, fs, session=None):
        """
        Initializes path, session and func to trigger cmd

        :param fs: proc/sys filesystem path
        :param session: ShellSession object of remote or VM
        """
        self.fs = fs
        self.func = process.getstatusoutput
        self.session = session
        if not self._check_isfile():
            raise AttributeError("%s is not available" % self.fs)
        if self.session:
            self.func = self.session.cmd_status_output

    def _check_isfile(self):
        """
        check whether the fs exist in local/remote host/VM
        """
        if self.session:
            return self.session.cmd_status("cat %s" % self.fs) == 0
        else:
            return os.path.isfile(self.fs)

    @property
    def fs_value(self):
        """
        Getter method for FS

        :return: String, current value in given filesystem
        """
        self.fs_val = str(self.func("cat %s" % self.fs)[1])
        return self.fs_val

    @fs_value.setter
    def fs_value(self, value):
        """
        Setter method for FS

        :param value: value to be set for given filesystem
        :return: Boolean, True on successfully set False on failure
        """
        # set the value
        cmd = "echo %s > %s" % (value, self.fs)
        status, output = self.func(cmd)
        if status != 0:
            logging.error("Failed to set %s to %s, error: %s", self.fs,
                          value, output.strip())
            return False
        return True


class ProcFS(FS):
    """
    class to get or set procfs values in host, remote host or in VM

    Example:
    >>> obj = ProcFS("/proc/sys/vm/nr_hugepages")
    >>> obj.proc_fs_value # To get the value
    >>> obj.proc_fs_value = 1 # To set the value
    """

    def __init__(self, proc_fs, session=None):
        """
        Initializes path, session and func to trigger cmd

        :param proc_fs: proc filesystem path
        :param session: ShellSession object of remote or VM
        """
        self.proc_fs = proc_fs
        super(ProcFS, self).__init__(self.proc_fs, session=session)

    @property
    def proc_fs_value(self):
        """
        Getter method for ProcFS

        :return: String, current value in given proc filesystem
        """
        try:
            return int(self.fs_value)
        except ValueError:
            return self.fs_val

    @proc_fs_value.setter
    def proc_fs_value(self, value):
        """
        Setter method for ProcFS

        :param value: value to be set for given proc filesystem
        :return: Boolean, True on successfully set False on failure
        """
        if str(value) != self.fs_value:
            self.fs_value = value
        # check if the value is reflected after set
        return str(value) == self.fs_value


class SysFS(FS):
    """
    class to get or set sysfs values in host, remote host or in VM

    Example:
    >>> obj = SysFS("/sys/kernel/mm/transparent_hugepage/enabled")
    >>> obj.sys_fs_value # To get the value
    >>> obj.sys_fs_value = "never" # To set the value
    """
    def __init__(self, sys_fs, session=None, regex=r"\[%s\]"):
        """
        Initializes path, session and func to trigger cmd

        :param sys_fs: sys filesystem path
        :param session: ShellSession object of remote or VM
        :param regex: delimiter to the sysfs enabled option.
        Example: # cat /sys/kernel/mm/transparent_hugepage/enabled
                 [always] madvise never
        """
        self.sys_fs = sys_fs
        self.regex = regex
        self.pattern = self.regex % ".*"
        super(SysFS, self).__init__(self.sys_fs, session=session)

    @property
    def sys_fs_value(self):
        """
        Getter method for SysFS

        :return: String, current value in given sys filesystem
        """
        output = re.search(self.pattern, self.fs_value)
        if output:
            return str(output.group()).strip()
        try:
            return int(self.fs_val)
        except ValueError:
            return self.fs_val

    @sys_fs_value.setter
    def sys_fs_value(self, value):
        """
        Setter method for SysFS

        :param value: value to be set for given sys filesystem
        :return: Boolean, True on successfully set False on failure
        """
        check_value = str(value)
        if re.search(self.pattern, str(self.sys_fs_value)):
            check_value = self.regex % value

        if check_value != self.fs_value:
            self.fs_value = value
            return check_value == self.fs_value
        return True



"""
Support for the pseudo 'emulator' device XML

http://libvirt.org/formatdomain.html#elementsDevices
"""

from virttest.libvirt_xml import accessors
from virttest.libvirt_xml.devices import base


class Emulator(base.UntypedDeviceBase):

    __slots__ = ('path',)

    def __init__(self, virsh_instance=base.base.virsh):
        accessors.XMLElementText('path', self, parent_xpath='/',
                                 tag_name='emulator')
        super(Emulator, self).__init__(device_tag='emulator',
                                       virsh_instance=virsh_instance)



"""
sound device support class(es)

http://libvirt.org/formatdomain.html#elementsSound
"""

from virttest.libvirt_xml import accessors
from virttest.libvirt_xml.devices import base


class Sound(base.UntypedDeviceBase):

    __slots__ = ('model_type', 'codec_type', 'address')

    def __init__(self, virsh_instance=base.base.virsh):
        accessors.XMLAttribute('model_type', self,
                               parent_xpath='/',
                               tag_name='sound',
                               attribute='model')
        accessors.XMLAttribute('codec_type', self,
                               parent_xpath='/',
                               tag_name='codec',
                               attribute='type')
        accessors.XMLElementDict('address', self,
                                 parent_xpath='/',
                                 tag_name='address')
        super(Sound, self).__init__(device_tag='sound',
                                    virsh_instance=virsh_instance)




"""
Virsh domjobinfo command related utility functions
"""

import re
import math
import logging

from avocado.core import exceptions

from virttest import virsh


# pylint: disable=E1121
def check_domjobinfo(vm, params, option="", remote_virsh_dargs=None):
    """
    Check given item in domjobinfo of the guest is as expected

    :param vm: the vm object
    :param params: the parameters used
    :param option: options for domjobinfo
    :param remote_virsh_dargs: the parameters for remote host
    :raise: exceptions.TestFail if the value of given item is unexpected
    """
    def _search_jobinfo(jobinfo, ignore_status=False):
        """
        Find value of given item in domjobinfo

        :param jobinfo: cmdResult object
        :param ignore_status: False to raise Error, True to ignore
        :raise: exceptions.TestFail if not found
        """
        for item in jobinfo.stdout.splitlines():
            if item.count(jobinfo_item):
                groups = re.findall(r'[0-9.]+', item.strip())
                logging.debug("In '%s' search '%s'\n", item, groups[0])
                if (math.fabs(float(groups[0]) - float(compare_to_value)) //
                   float(compare_to_value) > diff_rate):
                    err_msg = ("{} {} has too much difference from "
                               "{}".format(jobinfo_item,
                                           groups[0],
                                           compare_to_value))
                    if ignore_status:
                        logging.error(err_msg)
                    else:
                        raise exceptions.TestFail(err_msg)
                break

    jobinfo_item = params.get("jobinfo_item")
    compare_to_value = params.get("compare_to_value")
    ignore_status = params.get("domjob_ignore_status", False)
    logging.debug("compare_to_value:%s", compare_to_value)
    diff_rate = float(params.get("diff_rate", "0"))
    if not jobinfo_item or not compare_to_value:
        return
    jobinfo = virsh.domjobinfo(vm.name, option, debug=True)
    _search_jobinfo(jobinfo, ignore_status)

    check_domjobinfo_remote = params.get("check_domjobinfo_remote")
    if check_domjobinfo_remote and remote_virsh_dargs:
        remote_virsh_session = virsh.VirshPersistent(**remote_virsh_dargs)
        jobinfo = remote_virsh_session.domjobinfo(vm.name, option, debug=True)
        _search_jobinfo(jobinfo, ignore_status)
        remote_virsh_session.close_session()



"""
Virtualization test - Virtual disk related utility functions

:copyright: Red Hat Inc.
"""
import os
import glob
import shutil
import stat
import random
import string
import tempfile
import logging
import re
try:
    import configparser as ConfigParser
except ImportError:
    import ConfigParser
from functools import cmp_to_key

from avocado.core import exceptions
from avocado.utils import process
from avocado.utils.service import SpecificServiceManager

from virttest import error_context
from virttest import utils_numeric
from virttest import remote

PARTITION_TABLE_TYPE_MBR = "msdos"
PARTITION_TABLE_TYPE_GPT = "gpt"
PARTITION_TYPE_PRIMARY = "primary"
PARTITION_TYPE_EXTENDED = "extended"
PARTITION_TYPE_LOGICAL = "logical"

SIZE_AVAILABLE = "available size"

# Whether to print all shell commands called
DEBUG = False


def copytree(src, dst, overwrite=True, ignore=''):
    """
    Copy dirs from source to target.

    :param src: source directory
    :param dst: destination directory
    :param overwrite: overwrite file if exist or not
    :param ignore: files want to ignore
    """
    ignore = glob.glob(os.path.join(src, ignore))
    for root, dirs, files in os.walk(src):
        dst_dir = root.replace(src, dst)
        if not os.path.exists(dst_dir):
            os.makedirs(dst_dir)
        for _ in files:
            if _ in ignore:
                continue
            src_file = os.path.join(root, _)
            dst_file = os.path.join(dst_dir, _)
            if os.path.exists(dst_file):
                if overwrite:
                    os.remove(dst_file)
                else:
                    continue
            shutil.copy(src_file, dst_dir)


def is_mount(src, dst=None, fstype=None, options=None, verbose=False,
             session=None):
    """
    Check is src or dst mounted.

    :param src: source device or directory
    :param dst: mountpoint, if None will skip to check
    :param fstype: file system type, if None will skip to check
    :param options: mount options should be seperated by ","
    :param session: check within the session if given

    :return: True if mounted, else return False
    """
    mount_str = "%s %s %s" % (src, dst, fstype)
    mount_str = mount_str.replace('None', '').strip()
    mount_list_cmd = 'cat /proc/mounts'

    if session:
        mount_result = session.cmd_output_safe(mount_list_cmd)
    else:
        mount_result = process.run(mount_list_cmd, shell=True).stdout_text
    if verbose:
        logging.debug("/proc/mounts contents:\n%s", mount_result)

    for result in mount_result.splitlines():
        if mount_str in result:
            if options:
                options = options.split(",")
                options_result = result.split()[3].split(",")
                for op in options:
                    if op not in options_result:
                        if verbose:
                            logging.info("%s is not mounted with given"
                                         " option %s", src, op)
                        return False
            if verbose:
                logging.info("%s is mounted", src)
            return True
    if verbose:
        logging.info("%s is not mounted", src)
    return False


def mount(src, dst, fstype=None, options=None, verbose=False, session=None):
    """
    Mount src under dst if it's really mounted, then remout with options.

    :param src: source device or directory
    :param dst: mountpoint
    :param fstype: filesystem type need to mount
    :param options: mount options
    :param session: mount within the session if given

    :return: if mounted return True else return False
    """
    options = (options and [options] or [''])[0]
    if is_mount(src, dst, fstype, options, verbose, session):
        if 'remount' not in options:
            options = 'remount,%s' % options
    cmd = ['mount']
    if fstype:
        cmd.extend(['-t', fstype])
    if options:
        cmd.extend(['-o', options])
    cmd.extend([src, dst])
    cmd = ' '.join(cmd)
    if session:
        return session.cmd_status(cmd, safe=True) == 0
    return process.system(cmd, verbose=verbose) == 0


def umount(src, dst, fstype=None, verbose=False, session=None):
    """
    Umount src from dst, if src really mounted under dst.

    :param src: source device or directory
    :param dst: mountpoint
    :param fstype: fstype used to check if mounted as expected
    :param session: umount within the session if given

    :return: if unmounted return True else return False
    """
    mounted = is_mount(src, dst, fstype, verbose=verbose, session=session)
    if mounted:
        from . import utils_package
        package = "psmisc"
        # check package is available, if not try installing it
        if not utils_package.package_install(package):
            logging.error("%s is not available/installed for fuser", package)
        fuser_cmd = "fuser -km %s" % dst
        umount_cmd = "umount %s" % dst
        if session:
            if not isinstance(session, remote.VMManager):
                session.cmd_output_safe(fuser_cmd)
            return session.cmd_status(umount_cmd, safe=True) == 0
        process.system(fuser_cmd, ignore_status=True, verbose=True, shell=True)
        return process.system(umount_cmd, ignore_status=True, verbose=True) == 0
    return True


@error_context.context_aware
def cleanup(folder):
    """
    If folder is a mountpoint, do what is possible to unmount it. Afterwards,
    try to remove it.

    :param folder: Directory to be cleaned up.
    """
    error_context.context(
        "cleaning up unattended install directory %s" % folder)
    umount(None, folder)
    if os.path.isdir(folder):
        shutil.rmtree(folder)


@error_context.context_aware
def clean_old_image(image):
    """
    Clean a leftover image file from previous processes. If it contains a
    mounted file system, do the proper cleanup procedures.

    :param image: Path to image to be cleaned up.
    """
    error_context.context("cleaning up old leftover image %s" % image)
    if os.path.exists(image):
        umount(image, None)
        os.remove(image)


def get_linux_disks(session, partition=False):
    """
    List all disks or disks with no partition.

    :param session: session object to guest
    :param partition: if true, list all disks; otherwise,
                      list only disks with no partition.
    :return: the disks dict.
             e.g. {kname: [kname, size, type, serial, wwn]}
    """
    list_disk_cmd = "lsblk -o KNAME,SIZE,TYPE,SERIAL,WWN"
    get_disk_name_cmd = "lsblk -no pkname /dev/%s"
    output = session.cmd_output(list_disk_cmd)
    devs = output.splitlines()
    disks_dict = {}
    part_dict = {}
    for dev in devs:
        dev = dev.split()
        if "disk" in dev:
            disks_dict[dev[0]] = dev
        if "part" in dev:
            part_dict[dev[0]] = dev
    if partition:
        disks_dict = dict(disks_dict, **part_dict)
    else:
        for part in part_dict.keys():
            output = session.cmd_output(get_disk_name_cmd % part)
            disk = output.splitlines()[0].strip()
            if disk in disks_dict.keys():
                disks_dict.pop(disk)
    return disks_dict


def get_windows_disks_index(session, image_size):
    """
    Get all disks index which show in 'diskpart list disk'.
    except for system disk.
    in diskpart: if disk size < 8GB: it displays as MB
                 else: it displays as GB

    :param session: session object to guest.
    :param image_size: image size. e.g. 40M
    :return: a list with all disks index except for system disk.
    """
    disk = "disk_" + ''.join(random.sample(string.ascii_letters + string.digits, 4))
    disk_indexs = []
    list_disk_cmd = "echo list disk > " + disk
    list_disk_cmd += " && echo exit >> " + disk
    list_disk_cmd += " && diskpart /s " + disk
    list_disk_cmd += " && del /f " + disk
    disks = session.cmd_output(list_disk_cmd)
    size_type = image_size[-1] + "B"
    if size_type == "MB":
        disk_size = image_size[:-1] + " MB"
    elif size_type == "GB" and int(image_size[:-1]) < 8:
        disk_size = str(int(image_size[:-1]) * 1024) + " MB"
    else:
        disk_size = image_size[:-1] + " GB"
    regex_str = 'Disk (\d+).*?%s.*?%s' % (disk_size, disk_size)
    for disk in disks.splitlines():
        if disk.startswith("  Disk"):
            o = re.findall(regex_str, disk, re.I | re.M)
            if o:
                disk_indexs.append(o[0])
    return disk_indexs


def _wrap_windows_cmd(cmd):
    """
    add header and footer for cmd in order to run it in diskpart tool.

    :param cmd: cmd to be wrapped.
    :return: wrapped cmd
    """
    disk = "disk_" + ''.join(random.sample(string.ascii_letters + string.digits, 4))
    cmd_header = "echo list disk > " + disk
    cmd_header += " && echo select disk %s >> " + disk
    cmd_footer = " echo exit >> " + disk
    cmd_footer += " && diskpart /s " + disk
    cmd_footer += " && del /f " + disk
    cmd_list = []
    for i in cmd.split(";"):
        i += " >> " + disk
        cmd_list.append(i)
    cmd = " && ".join(cmd_list)
    return " && ".join([cmd_header, cmd, cmd_footer])


def update_windows_disk_attributes(session, dids, timeout=120):
    """
    Clear readonly for all disks and online them in windows guest.
    It is a workaround update attributes for all disks in one time.
    If configures disk(update attribute -> format) one by one,
    it will hit error.

    :param session: session object to guest.
    :param dids: a list with all disks index which
                 show in 'diskpart list disk'.
                 call function: get_windows_disks_index()
    :param timeout: time for cmd execution
    :return: True or False
    """
    detail_cmd = ' echo detail disk'
    detail_cmd = _wrap_windows_cmd(detail_cmd)
    set_rw_cmd = ' echo attributes disk clear readonly'
    set_rw_cmd = _wrap_windows_cmd(set_rw_cmd)
    online_cmd = ' echo online disk'
    online_cmd = _wrap_windows_cmd(online_cmd)
    for did in dids:
        logging.info("Detail for 'Disk%s'" % did)
        details = session.cmd_output(detail_cmd % did)
        if re.search("Read.*Yes", details, re.I | re.M):
            logging.info("Clear readonly bit on 'Disk%s'" % did)
            status, output = session.cmd_status_output(set_rw_cmd % did,
                                                       timeout=timeout)
            if status != 0:
                logging.error("Can not clear readonly bit: %s" % output)
                return False
        if re.search("Status.*Offline", details, re.I | re.M):
            logging.info("Online 'Disk%s'" % did)
            status, output = session.cmd_status_output(online_cmd % did,
                                                       timeout=timeout)
            if status != 0:
                logging.error("Can not online disk: %s" % output)
                return False
    return True


def create_partition_table_linux(session, did, labeltype):
    """
    Create partition table on disk in linux guest.

    :param session: session object to guest.
    :param did: disk kname. e.g. sdb
    :param labeltype: label type for the disk.
    """
    mklabel_cmd = 'parted -s "/dev/%s" mklabel %s' % (did, labeltype)
    session.cmd(mklabel_cmd)


def create_partition_table_windows(session, did, labeltype):
    """
    Create partition table on disk in windows guest.
    mbr is default value, do nothing.

    :param session: session object to guest.
    :param did: disk index which show in 'diskpart list disk'.
    :param labeltype: label type for the disk.
    """
    if labeltype == PARTITION_TABLE_TYPE_GPT:
        mklabel_cmd = ' echo convert gpt'
        mklabel_cmd = _wrap_windows_cmd(mklabel_cmd)
        session.cmd(mklabel_cmd % did)


def create_partition_table(session, did, labeltype, ostype):
    """
    Create partition table on disk in linux or windows guest.

    :param session: session object to guest.
    :param did: disk kname or disk index
    :param labeltype: label type for the disk.
    :param ostype: linux or windows.
    """
    if ostype == "windows":
        create_partition_table_windows(session, did, labeltype)
    else:
        create_partition_table_linux(session, did, labeltype)


def create_partition_linux(session, did, size, start,
                           part_type=PARTITION_TYPE_PRIMARY, timeout=360):
    """
    Create single partition on disk in linux guest.

    :param session: session object to guest.
    :param did: disk kname. e.g. sdb
    :param size: partition size. e.g. 200M
    :param start: partition beginning at start. e.g. 0M
    :param part_type: partition type, primary extended logical
    :param timeout: Timeout for cmd execution in seconds.
    """
    size = utils_numeric.normalize_data_size(size, order_magnitude="M") + "M"
    start = utils_numeric.normalize_data_size(start, order_magnitude="M") + "M"
    end = str(float(start[:-1]) + float(size[:-1])) + size[-1]
    partprobe_cmd = "partprobe /dev/%s" % did
    mkpart_cmd = 'parted -s "%s" mkpart %s %s %s'
    mkpart_cmd %= ("/dev/%s" % did, part_type, start, end)
    session.cmd(mkpart_cmd)
    session.cmd(partprobe_cmd, timeout=timeout)


def create_partition_windows(session, did, size, start,
                             part_type=PARTITION_TYPE_PRIMARY, timeout=360):
    """
    Create single partition on disk in windows guest.

    :param session: session object to guest.
    :param did: disk index
    :param size: partition size. e.g. size 200M
    :param start: partition beginning at start. e.g. 0M
    :param part_type: partition type, primary extended logical
    :param timeout: Timeout for cmd execution in seconds.
    """
    size = utils_numeric.normalize_data_size(size, order_magnitude="M")
    start = utils_numeric.normalize_data_size(start, order_magnitude="M")
    size = int(float(size) - float(start))
    mkpart_cmd = " echo create partition %s size=%s"
    mkpart_cmd = _wrap_windows_cmd(mkpart_cmd)
    session.cmd(mkpart_cmd % (did, part_type, size), timeout=timeout)


def create_partition(session, did, size, start, ostype,
                     part_type=PARTITION_TYPE_PRIMARY, timeout=360):
    """
    Create single partition on disk in windows or linux guest.

    :param session: session object to guest.
    :param did: disk kname or disk index
    :param size: partition size. e.g. size 2G
    :param start: partition beginning at start. e.g. 0G
    :param ostype: linux or windows.
    :param part_type: partition type, primary extended logical
    :param timeout: Timeout for cmd execution in seconds.
    """
    if ostype == "windows":
        create_partition_windows(session, did, size, start, part_type, timeout)
    else:
        create_partition_linux(session, did, size, start, part_type, timeout)


def delete_partition_linux(session, partition_name, timeout=360):
    """
    remove single partition for one disk.

    :param session: session object to guest.
    :param partition_name: partition name. e.g. sdb1
    :param timeout: Timeout for cmd execution in seconds.
    """
    get_kname_cmd = "lsblk -no pkname /dev/%s"
    kname = session.cmd_output(get_kname_cmd % partition_name)
    list_disk_cmd = "lsblk -o KNAME,MOUNTPOINT"
    output = session.cmd_output(list_disk_cmd)
    output = output.splitlines()
    rm_cmd = 'parted -s "/dev/%s" rm %s'
    for line in output:
        partition = re.findall(partition_name, line, re.I | re.M)
        if partition:
            if "/" in line.split()[-1]:
                if not umount("/dev/%s" % partition_name, line.split()[-1], session=session):
                    err_msg = "Failed to umount partition '%s'"
                    raise exceptions.TestError(err_msg % partition_name)
            break
    session.cmd(rm_cmd % (kname, partition[0]))
    session.cmd("partprobe /dev/%s" % kname, timeout=timeout)


def delete_partition_windows(session, partition_name, timeout=360):
    """
    remove single partition for one disk.

    :param session: session object to guest.
    :param partition_name: partition name. e.g. D
    :param timeout: Timeout for cmd execution in seconds.
    """
    disk = "disk_" + ''.join(random.sample(string.ascii_letters + string.digits, 4))
    delete_cmd = "echo select volume %s > " + disk
    delete_cmd += " && echo delete volume >> " + disk
    delete_cmd += " && echo exit >> " + disk
    delete_cmd += " && diskpart /s " + disk
    delete_cmd += " && del /f " + disk
    session.cmd(delete_cmd % partition_name, timeout=timeout)


def delete_partition(session, partition_name, ostype, timeout=360):
    """
    remove single partition for one disk in linux or windows guest.

    :param session: session object to guest.
    :param partition_name: partition name.
    :param ostype: linux or windows.
    :param timeout: Timeout for cmd execution in seconds.
    """
    if ostype == "windows":
        delete_partition_windows(session, partition_name, timeout)
    else:
        delete_partition_linux(session, partition_name, timeout)


def clean_partition_linux(session, did, timeout=360):
    """
    clean partition for linux guest.

    :param session: session object to guest.
    :param did: disk ID in guest.
                for linux: disk kname. e.g. 'sdb', 'nvme0n1'
    :param timeout: Timeout for cmd execution in seconds.
    """
    list_disk_cmd = "lsblk -o KNAME,MOUNTPOINT"
    output = session.cmd_output(list_disk_cmd)
    output = output.splitlines()
    regex_str = did + "\w*(\d+)"
    rm_cmd = 'parted -s "/dev/%s" rm %s'
    for line in output:
        partition = re.findall(regex_str, line, re.I | re.M)
        if partition:
            if "/" in line.split()[-1]:
                if not umount("/dev/%s" % line.split()[0], line.split()[-1], session=session):
                    err_msg = "Failed to umount partition '%s'"
                    raise exceptions.TestError(err_msg % line.split()[0])
    list_partition_number = "parted -s /dev/%s print|awk '/^ / {print $1}'"
    partition_numbers = session.cmd_output(list_partition_number % did)
    ignore_err_msg = "unrecognised disk label"
    if ignore_err_msg in partition_numbers:
        logging.info("no partition to clean on %s" % did)
    else:
        partition_numbers = partition_numbers.splitlines()
        for number in partition_numbers:
            logging.info("remove partition %s on %s" % (number, did))
            session.cmd(rm_cmd % (did, number))
        session.cmd("partprobe /dev/%s" % did, timeout=timeout)


def clean_partition_windows(session, did, timeout=360):
    """
    clean partition for windows guest.

    :param session: session object to guest.
    :param did: disk ID in guest.
                for windows: disk index which
                             show in 'diskpart list disk'.
                             e.g. 1, 2
    :param timeout: Timeout for cmd execution in seconds.
    """
    clean_cmd = " echo clean "
    clean_cmd = _wrap_windows_cmd(clean_cmd)
    session.cmd(clean_cmd % did, timeout=timeout)


def clean_partition(session, did, ostype, timeout=360):
    """
    clean partition for disk in linux or windows guest.

    :param session: session object to guest.
    :param did: disk ID in guest.
                for linux: disk kname. e.g. 'sdb', 'sdc'
                for windows: disks index which
                             show in 'diskpart list disk'.
                             e.g. 1, 2
    :param ostype: guest os type 'windows' or 'linux'.
    :param timeout: Timeout for cmd execution in seconds.
    """
    if ostype == "windows":
        clean_partition_windows(session, did, timeout)
    else:
        clean_partition_linux(session, did, timeout)


def create_filesyetem_linux(session, partition_name, fstype, timeout=360):
    """
    create file system in linux guest.

    :param session: session object to guest.
    :param partition_name: partition name that to be formatted. e.g. sdb1
    :param fstype: filesystem type for the disk.
    :param timeout: Timeout for cmd execution in seconds.
    """
    if fstype == "xfs":
        mkfs_cmd = "mkfs.%s -f" % fstype
    else:
        mkfs_cmd = "mkfs.%s -F" % fstype
    format_cmd = "yes|%s '/dev/%s'" % (mkfs_cmd, partition_name)
    session.cmd(format_cmd, timeout=timeout)


def create_filesystem_windows(session, partition_name, fstype,
                              timeout=360, quick_format=True):
    """
    create file system in windows guest.

    :param session: session object to guest.
    :param partition_name: partition name that to be formatted. e.g. D
    :param fstype: file system type for the disk.
    :param timeout: Timeout for cmd execution in seconds.
    :param quick_format: Whether use quick format or not.
    """
    disk = "disk_" + ''.join(random.sample(string.ascii_letters + string.digits, 4))
    format_cmd = "echo select volume %s > " + disk
    format_cmd += " && echo format fs=%s "
    if quick_format:
        format_cmd += "quick "
    format_cmd += ">> " + disk
    format_cmd += " && echo exit >> " + disk
    format_cmd += " && diskpart /s " + disk
    format_cmd += " && del /f " + disk
    session.cmd(format_cmd % (partition_name, fstype), timeout=timeout)


def create_filesystem(session, partition_name, fstype, ostype, timeout=360):
    """
    create file system in windows or linux guest.

    :param session: session object to guest.
    :param partition_name: partition name that to be formatted.
    :param fstype: file system type for the disk.
    :param ostype: guest os type 'windows' or 'linux'.
    :param timeout: Timeout for cmd execution in seconds.
    """
    if ostype == "windows":
        create_filesystem_windows(session, partition_name, fstype, timeout)
    else:
        create_filesyetem_linux(session, partition_name, fstype, timeout)


def _get_mpoint_fstype_linux(session, partition):
    """
    Get a partition's file system type and mountpoint.

    :param session: session object to guest.
    :param partition: disk partition, like /dev/sdb1.
    :return: Tuple (mountpoint, fstype)
    """
    mount_list = session.cmd_output('cat /proc/mounts')
    mount_info = re.search(r'%s\s(.+?)\s(.+?)\s' % partition, mount_list)
    return mount_info.groups()


def get_scsi_info(device_source):
    """
    Gets scsi info

    :param device_source: source device path, e.g. /dev/sda
    :raise TestError: if 'lsscsi' output is unexpected
    :return: "scsi_num:bus_num:target_num:unit_num"
    """
    cmd = "lsscsi | grep %s | awk '{print $1}'" % device_source
    cmd_result = process.run(cmd, shell=True)
    scsi_info = re.findall("\d+", str(cmd_result.stdout.strip()))
    if len(scsi_info) != 4:
        raise exceptions.TestError("Got wrong scsi info: %s" % scsi_info)
    return ":".join(scsi_info)


def get_partition_attrs_linux(session, partition):
    """
    Get partition attributes in linux guest.

    :param session: session object to guest.
    :param partition: like /dev/sdb1.
    :return: dict like {'start': '512B', 'end': '16106127359B',
                       'size': '16106126848B', 'type': 'primary'}
    """
    pattern = r'(/dev/.*)p(\d+)' if "nvme" in partition else r'(/dev/.*)(\d+)'
    dev_name, part_num = re.match(pattern, partition).groups()
    parted_cmd = 'parted -s %s unit B print' % dev_name
    pattern = re.compile(r'%s\s+(?P<start>\d+\w+)\s+(?P<end>\d+\w+)\s+'
                         r'(?P<size>\d+\w+)\s+(?P<type>\w+)' % part_num)
    return pattern.search(session.cmd(parted_cmd), re.M).groupdict()


def resize_filesystem_linux(session, partition, size):
    """
    Resize file system in linux guest.
    For ext2, ext3, ext4 filesystem, support enlarge and shrink.
    For xfs filesystem, only support enlarge, not support shrink.

    :param session: session object to guest.
    :param partition: disk partition, like /dev/sdb1.
    :param size: resize file system to size.
                 size unit can be 'B', 'K', 'M', 'G'.
                 support transfer size with SIZE_AVAILABLE,
                 enlarge to maximun available size.
    """
    def get_start_size():
        start_size = get_partition_attrs_linux(session, partition)['start']
        return int(utils_numeric.normalize_data_size(start_size, 'B').split('.')[0])

    def resize_xfs_fs(size):
        if size == SIZE_AVAILABLE:
            resize_fs_cmd = 'xfs_growfs -d %s' % mountpoint
        else:
            output = session.cmd_output('xfs_growfs -n %s' % mountpoint)
            bsize = int(re.findall(r'data\s+=\s+bsize=(\d+)', output, re.M)[0])
            blocks = (int(utils_numeric.normalize_data_size(size, 'B').split('.')[0]) -
                      get_start_size()) // bsize
            resize_fs_cmd = 'xfs_growfs -D %s %s' % (blocks, mountpoint)
        session.cmd(resize_fs_cmd)

    def resize_ext_fs(size):
        flag = False
        if is_mount(partition, dst=mountpoint, fstype=fstype, session=session):
            umount(partition, mountpoint, fstype=fstype, session=session)
            flag = True

        session.cmd('e2fsck -f %s' % partition)

        if size == SIZE_AVAILABLE:
            resize_fs_cmd = 'resize2fs %s' % partition
        else:
            output = session.cmd_output('tune2fs -l %s | grep -i block' % partition)
            bsize = int(re.findall(r'Block size:\s+(\d+)', output, re.M)[0])
            size = ((int(utils_numeric.normalize_data_size(size, 'B').split(".")[0]) -
                     get_start_size()) // bsize) * bsize
            size = utils_numeric.normalize_data_size(str(size).split(".")[0], 'K')
            resize_fs_cmd = 'resize2fs %s %sK' % (partition, int(size.split(".")[0]))
        session.cmd(resize_fs_cmd)
        if flag:
            mount(partition, mountpoint, fstype=fstype, session=session)

    mountpoint, fstype = _get_mpoint_fstype_linux(session, partition)
    if fstype == 'xfs':
        resize_xfs_fs(size)
    elif fstype.startswith("ext"):
        resize_ext_fs(size)
    else:
        raise NotImplementedError


def resize_partition_linux(session, partition, size):
    """
    Resize partition in linux guest.
    Note: not support gpt type resize for linux guest.

    :param session: session object to guest.
    :param partition: partition that to be shrunk, like /dev/sdb1.
    :param size: resize partition to size, unit is B.
    """
    mountpoint, fstype = _get_mpoint_fstype_linux(session, partition)
    flag = False
    if is_mount(partition, dst=mountpoint, fstype=fstype, session=session):
        umount(partition, mountpoint, fstype=fstype, session=session)
        flag = True

    # FIXME: if nvme device need support this function.
    dev_name, part_num = re.match(r'(/dev/.*)(\d+)', partition).groups()
    parted_cmd = 'parted -s %s print' % dev_name
    part_attrs = get_partition_attrs_linux(session, partition)
    start_size = int(utils_numeric.normalize_data_size(part_attrs['start'],
                                                       'B').split('.')[0])
    end_size = (int(utils_numeric.normalize_data_size(size, 'B').split('.')[0])
                - start_size)
    session.cmd(' '.join((parted_cmd, 'rm %s' % part_num)))
    resizepart_cmd = ' '.join((parted_cmd, 'unit B mkpart {0} {1} {2}'))
    session.cmd(resizepart_cmd.format(part_attrs['type'], start_size, end_size))
    session.cmd('partprobe %s' % dev_name)

    if fstype == 'xfs':
        session.cmd('xfs_repair -n %s' % partition)
    elif fstype.startswith("ext"):
        session.cmd('e2fsck -f %s' % partition)
    else:
        raise NotImplementedError

    if flag:
        mount(partition, mountpoint, fstype=fstype, session=session)


def get_disk_size_windows(session, did):
    """
    Get disk size from windows guest.

    :param session: session object to guest.
    :param did: disk index which show in 'diskpart list disk'.
                e.g. 0, 1
    :return: disk size.
    """
    cmd = "wmic diskdrive get size, index"
    return int(re.findall(r'%s\s+(\d+)' % did, session.cmd_output(cmd))[0])


def get_disk_size_linux(session, did):
    """
    Get disk size from linux guest.

    :param session: session object to guest.
    :param did: disk kname. e.g. 'sdb', 'sdc'
    :return: disk size.
    """
    disks_info = get_linux_disks(session, partition=True)
    disk_size = disks_info['%s' % did][1]
    return int(utils_numeric.normalize_data_size(disk_size, 'B').split('.')[0])


def get_disk_size(session, os_type, did):
    """
    Get disk size from guest.

    :param session: session object to guest.
    :param ostype: guest os type 'windows' or 'linux'.
    :param did: disk ID in guest.
    :return: disk size.
    """
    if os_type == "linux":
        return get_disk_size_linux(session, did)
    else:
        return get_disk_size_windows(session, did)


def get_drive_letters(session, did):
    """
    Get drive letters of specified disk in windows guest.

    :param session: Windows VM session.
    :type session: aexpect.ShellSession
    :param did: The disk index.
    :type did: str
    :return: The drive letters.
    :rtype: list
    """
    disk_script = "disk_" + ''.join(
        random.sample(string.ascii_letters + string.digits, 4))
    select_cmd = "echo select disk %s > %s " % (did, disk_script)
    detail_cmd = "echo detail disk >> %s" % disk_script
    diskpart_cmd = "diskpart /s %s" % disk_script
    cmd = ' && '.join((select_cmd, detail_cmd, diskpart_cmd))
    output = session.cmd(cmd)

    letter_offset = 0
    searched_ret = re.search(r"(.*Volume\s+.*Ltr.*)", output, re.I | re.M)
    if searched_ret:
        letter_offset = searched_ret.group(1).index('Ltr') + 1

    if letter_offset:
        vol_details = re.findall(r"(.*Volume\s+\d+.*Partition.*)", output, re.I | re.M)
        return [_[letter_offset] for _ in vol_details if _[letter_offset].isalpha()]
    return []


def set_drive_letter(session, did, partition_no=1):
    """
    set drive letter for partition in windows guest.

    :param session: session object to guest.
    :param did: disk index
    :param partition_no: partition number
    :return drive_letter: drive letter has been set
    """
    drive_letter = ""
    list_partition_cmd = ' echo list partition '
    list_partition_cmd = _wrap_windows_cmd(list_partition_cmd)
    details = session.cmd_output(list_partition_cmd % did)
    for line in details.splitlines():
        if re.search("Reserved", line, re.I | re.M):
            partition_no += int(line.split()[1])
    assign_letter_cmd = ' echo select partition %s ; echo assign '
    assign_letter_cmd = _wrap_windows_cmd(assign_letter_cmd)
    session.cmd(assign_letter_cmd % (did, partition_no))
    detail_cmd = ' echo detail disk '
    detail_cmd = _wrap_windows_cmd(detail_cmd)
    details = session.cmd_output(detail_cmd % did)
    for line in details.splitlines():
        pattern = "\s+Volume\s+\d+"
        if re.search(pattern, line, re.I | re.M):
            drive_letter = line.split()[2]
            break
    if len(drive_letter) == 1 and drive_letter.isalpha():
        return drive_letter
    else:
        return None


def drop_drive_letter(session, drive_letter):
    """
    remove drive letter for partition in windows guest.

    :param session: session object to guest.
    :param drive_letter: drive letter to be removed
    """
    disk = "disk_" + ''.join(random.sample(string.ascii_letters + string.digits, 4))
    remove_cmd = "echo select volume %s > " + disk
    remove_cmd += " && echo remove >> " + disk
    remove_cmd += " && echo exit >> " + disk
    remove_cmd += " && diskpart /s " + disk
    remove_cmd += " && del /f " + disk
    session.cmd(remove_cmd % drive_letter)


def configure_empty_windows_disk(session, did, size, start="0M",
                                 n_partitions=1, fstype="ntfs",
                                 labeltype=PARTITION_TABLE_TYPE_MBR,
                                 timeout=360, quick_format=True):
    """
    Create partition on disks in windows guest, format and mount it.
    Only handle an empty disk and will create equal size partitions onto the disk.

    :param session: session object to guest.
    :param did: disk index which show in 'diskpart list disk'.
    :param size: partition size. e.g. 500M
    :param start: partition beginning at start. e.g. 0M
    :param n_partitions: the number of partitions on disk
    :param fstype: filesystem type for the disk.
    :param labeltype: label type for the disk.
    :param timeout: Timeout for cmd execution in seconds.
    :param quick_format: Whether use quick format or not.
    :return a list: mount point list for all partitions.
    """
    mountpoint = []
    create_partition_table_windows(session, did, labeltype)
    start = utils_numeric.normalize_data_size(start, order_magnitude="M") + "M"
    partition_size = float(size[:-1]) / n_partitions
    extended_size = float(size[:-1]) - partition_size
    reserved_size = 5
    if labeltype == PARTITION_TABLE_TYPE_MBR and n_partitions > 1:
        part_type = PARTITION_TYPE_EXTENDED
    else:
        part_type = PARTITION_TYPE_PRIMARY
    for i in range(n_partitions):
        if i == 0:
            create_partition_windows(
                session, did, str(partition_size) + size[-1],
                str(float(start[:-1]) + reserved_size) + start[-1], timeout=timeout)
        else:
            if part_type == PARTITION_TYPE_EXTENDED:
                create_partition_windows(
                    session, did, str(extended_size) + size[-1], start, part_type, timeout)
                part_type = PARTITION_TYPE_LOGICAL
                create_partition_windows(
                    session, did, str(partition_size) + size[-1], start, part_type, timeout)
            else:
                create_partition_windows(
                    session, did, str(partition_size) + size[-1], start, part_type, timeout)
        drive_letter = set_drive_letter(session, did, partition_no=i + 1)
        if not drive_letter:
            return []
        mountpoint.append(drive_letter)
        create_filesystem_windows(session, mountpoint[i], fstype, timeout, quick_format=quick_format)
    return mountpoint


def configure_empty_linux_disk(session, did, size, start="0M", n_partitions=1,
                               fstype="ext4", labeltype=PARTITION_TABLE_TYPE_MBR,
                               timeout=360):
    """
    Create partition on disk in linux guest, format and mount it.
    Only handle an empty disk and will create equal size partitions onto the disk.
    Note: Make sure '/mnt/' is not mount point before run this function,
          in order to avoid unknown exceptions.

    :param session: session object to guest.
    :param did: disk kname. e.g. sdb
    :param size: partition size. e.g. 2G
    :param start: partition beginning at start. e.g. 0G
    :param n_partitions: the number of partitions on disk
    :param fstype: filesystem type for the disk.
    :param labeltype: label type for the disk.
    :param timeout: Timeout for cmd execution in seconds.
    :return a list: mount point list for all partitions.
    """
    mountpoint = []
    create_partition_table_linux(session, did, labeltype)
    size = utils_numeric.normalize_data_size(size, order_magnitude="M") + "M"
    start = float(utils_numeric.normalize_data_size(start, order_magnitude="M"))
    partition_size = float(size[:-1]) / n_partitions
    extended_size = float(size[:-1]) - partition_size
    if labeltype == PARTITION_TABLE_TYPE_MBR and n_partitions > 1:
        part_type = PARTITION_TYPE_EXTENDED
    else:
        part_type = PARTITION_TYPE_PRIMARY
    for i in range(n_partitions):
        pre_partition = get_linux_disks(session, partition=True).keys()
        if i == 0:
            create_partition_linux(session, did, str(partition_size) + size[-1],
                                   str(start) + size[-1], timeout=timeout)
        else:
            if part_type == PARTITION_TYPE_EXTENDED:
                create_partition_linux(session, did, str(extended_size) + size[-1],
                                       str(start) + size[-1], part_type, timeout)
                pre_partition = get_linux_disks(session, partition=True).keys()
                part_type = PARTITION_TYPE_LOGICAL
                create_partition_linux(session, did, str(partition_size) + size[-1],
                                       str(start) + size[-1], part_type, timeout)
            else:
                create_partition_linux(session, did, str(partition_size) + size[-1],
                                       str(start) + size[-1], part_type, timeout)
        start += partition_size
        post_partition = get_linux_disks(session, partition=True).keys()
        new_partition = list(set(post_partition) - set(pre_partition))[0]
        create_filesyetem_linux(session, new_partition, fstype, timeout)
        mount_dst = "/mnt/" + new_partition
        session.cmd("rm -rf %s; mkdir %s" % (mount_dst, mount_dst))
        if not mount("/dev/%s" % new_partition, mount_dst, fstype=fstype, session=session):
            err_msg = "Failed to mount partition '%s'"
            raise exceptions.TestError(err_msg % new_partition)
        mountpoint.append(mount_dst)
    return mountpoint


def configure_empty_disk(session, did, size, ostype, start="0M", n_partitions=1,
                         fstype=None, labeltype=PARTITION_TABLE_TYPE_MBR,
                         timeout=360):
    """
    Create partition on disk in guest, format and mount it.
    Only handle an empty disk and will create equal size partitions onto the disk.

    :param session: session object to guest.
    :param did: disk ID list in guest.
                for linux: disk kname, serial or wwn.
                           e.g. 'sdb'
                for windows: disk index which show in 'diskpart list disk'
                             call function: get_windows_disks_index()
    :param size: partition size. e.g. size 2G
    :param ostype: guest os type 'windows' or 'linux'.
    :param start: partition beginning at start. e.g. 0G
    :param n_partitions: the number of partitions on disk for guest
    :param fstype: filesystem type for the disk; when it's the default None,
                   it will use the default one for corresponding ostype guest
    :param labeltype: label type for the disk.
    :param timeout: Timeout for cmd execution in seconds.
    :return a list: mount point list for all partitions.
    """
    default_fstype = "ntfs" if (ostype == "windows") else "ext4"
    fstype = fstype or default_fstype
    if ostype == "windows":
        return configure_empty_windows_disk(session, did, size, start,
                                            n_partitions, fstype,
                                            labeltype, timeout)
    return configure_empty_linux_disk(session, did, size, start,
                                      n_partitions, fstype,
                                      labeltype, timeout)


def linux_disk_check(session, did):
    """
    Check basic functions for linux guest disk on localhost.
    Create partition on disk, format, mount, create a file and then unmount, clean
    partition

    :param session: VM session
    :param did: Disk Kname in VM. eg. vdb
    """

    mount_point = configure_empty_linux_disk(session, did, "100M")
    try:
        if len(mount_point) != 1:
            raise exceptions.TestError("Incorrect mount point {}".format(mount_point))
        cmd = "echo teststring >  {}/testfile".format(mount_point[0])
        if session.cmd_status(cmd):
            raise exceptions.TestError("Failed to run {}".format(cmd))
    finally:
        clean_partition_linux(session, did)


def get_parts_list(session=None):
    """
    Get all partition lists.
    """
    parts_cmd = "cat /proc/partitions"
    if session:
        _, parts_out = session.cmd_status_output(parts_cmd)
    else:
        parts_out = process.run(parts_cmd).stdout_text
    parts = []
    if parts_out:
        for line in parts_out.rsplit("\n"):
            if line.startswith("major") or line == "":
                continue
            parts_line = line.rsplit()
            if len(parts_line) == 4:
                parts.append(parts_line[3])
    logging.debug("Find parts: %s", parts)
    return parts


def get_disk_by_serial(serial_str, session=None):
    """
    Get disk by serial in VM or host

    :param serial_str: ID_SERIAL of disk, string value
    :param session: VM session or None
    :return: Disk name if find one with serial_str, else None
    """
    parts_list = get_parts_list(session=session)
    for disk in parts_list:
        cmd = ("udevadm info --query=all --name=/dev/{} | grep ID_SERIAL={}"
               .format(disk, serial_str))
        if session:
            status = session.cmd_status(cmd)
        else:
            status = process.run(cmd, shell=True, ignore_status=True).exit_status
        if not status:
            logging.debug("Disk %s has serial %s", disk, serial_str)
            return disk


def check_remote_vm_disks(params):
    """
    Check disks in remote vm are working well with I/O.

    :param params: the dict used for parameters
    """
    remote_vm_obj = remote.VMManager(params)
    remote_vm_obj.check_network()
    remote_vm_obj.setup_ssh_auth()
    disks = get_linux_disks(remote_vm_obj, False)
    logging.debug("Get disks in remote VM: %s", disks)

    for disk in disks.keys():
        linux_disk_check(remote_vm_obj, disk)


def dd_data_to_vm_disk(session, disk, bs='1M', seek='0', count='100'):
    """
    Generate some random data to a vm disk

    :param session: The vm session we'll use
    :param disk: The disk of the vm we'll use
    :param bs: The 'bs' param of the 'dd' command
    :param seek: The 'seek' param of the 'dd' command
    :param count: The 'count' param of the 'dd' command
    """
    dd_cmd = "dd if=/dev/urandom of=%s bs=%s seek=%s count=%s; sync"
    dd_cmd %= (disk, bs, seek, count)
    output = session.cmd_output(dd_cmd).strip()
    logging.debug("Using dd to generate data to %s: %s", disk, output)


class Disk(object):

    """
    Abstract class for Disk objects, with the common methods implemented.
    """

    def __init__(self):
        self.path = None

    def get_answer_file_path(self, filename):
        return os.path.join(self.mount, filename)

    def copy_to(self, src):
        logging.debug("Copying %s to disk image mount", src)
        dst = os.path.join(self.mount, os.path.basename(src))
        if os.path.isdir(src):
            shutil.copytree(src, dst)
        elif os.path.isfile(src):
            shutil.copyfile(src, dst)

    def close(self):
        os.chmod(self.path, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP |
                 stat.S_IROTH | stat.S_IXOTH)
        cleanup(self.mount)
        logging.debug("Disk %s successfully set", self.path)


class FloppyDisk(Disk):

    """
    Represents a floppy disk. We can copy files to it, and setup it in
    convenient ways.
    """
    @error_context.context_aware
    def __init__(self, path, qemu_img_binary, tmpdir, vfd_size):
        error_context.context(
            "Creating unattended install floppy image %s" % path)
        self.mount = tempfile.mkdtemp(prefix='floppy_virttest_', dir=tmpdir)
        self.path = path
        self.vfd_size = vfd_size
        clean_old_image(path)
        try:
            c_cmd = '%s create -f raw %s %s' % (qemu_img_binary, path,
                                                self.vfd_size)
            process.run(c_cmd, verbose=DEBUG)
            f_cmd = 'mkfs.msdos -s 1 %s' % path
            process.run(f_cmd, verbose=DEBUG)
        except process.CmdError as e:
            logging.error("Error during floppy initialization: %s" % e)
            cleanup(self.mount)
            raise

    def close(self):
        """
        Copy everything that is in the mountpoint to the floppy.
        """
        pwd = os.getcwd()
        try:
            os.chdir(self.mount)
            path_list = glob.glob('*')
            for path in path_list:
                self.copy_to(path)
        finally:
            os.chdir(pwd)

        cleanup(self.mount)

    def copy_to(self, src):
        logging.debug("Copying %s to floppy image", src)
        mcopy_cmd = "mcopy -s -o -n -i %s %s ::/" % (self.path, src)
        process.run(mcopy_cmd, verbose=DEBUG)

    def _copy_virtio_drivers(self, virtio_floppy):
        """
        Copy the virtio drivers on the virtio floppy to the install floppy.

        1) Mount the floppy containing the viostor drivers
        2) Copy its contents to the root of the install floppy
        """
        pwd = os.getcwd()
        try:
            m_cmd = 'mcopy -s -o -n -i %s ::/* %s' % (
                virtio_floppy, self.mount)
            process.run(m_cmd, verbose=DEBUG)
        finally:
            os.chdir(pwd)

    def setup_virtio_win2003(self, virtio_floppy, virtio_oemsetup_id):
        """
        Setup the install floppy with the virtio storage drivers, win2003 style.

        Win2003 and WinXP depend on the file txtsetup.oem file to install
        the virtio drivers from the floppy, which is a .ini file.
        Process:

        1) Copy the virtio drivers on the virtio floppy to the install floppy
        2) Parse the ini file with config parser
        3) Modify the identifier of the default session that is going to be
           executed on the config parser object
        4) Re-write the config file to the disk
        """
        self._copy_virtio_drivers(virtio_floppy)
        txtsetup_oem = os.path.join(self.mount, 'txtsetup.oem')

        if not os.path.isfile(txtsetup_oem):
            raise IOError('File txtsetup.oem not found on the install '
                          'floppy. Please verify if your floppy virtio '
                          'driver image has this file')

        parser = ConfigParser.ConfigParser()
        parser.read(txtsetup_oem)

        if not parser.has_section('Defaults'):
            raise ValueError('File txtsetup.oem does not have the session '
                             '"Defaults". Please check txtsetup.oem')

        default_driver = parser.get('Defaults', 'SCSI')
        if default_driver != virtio_oemsetup_id:
            parser.set('Defaults', 'SCSI', virtio_oemsetup_id)
            fp = open(txtsetup_oem, 'w')
            parser.write(fp)
            fp.close()

    def setup_virtio_win2008(self, virtio_floppy):
        """
        Setup the install floppy with the virtio storage drivers, win2008 style.

        Win2008, Vista and 7 require people to point out the path to the drivers
        on the unattended file, so we just need to copy the drivers to the
        driver floppy disk. Important to note that it's possible to specify
        drivers from a CDROM, so the floppy driver copy is optional.
        Process:

        1) Copy the virtio drivers on the virtio floppy to the install floppy,
           if there is one available
        """
        if os.path.isfile(virtio_floppy):
            self._copy_virtio_drivers(virtio_floppy)
        else:
            logging.debug(
                "No virtio floppy present, not needed for this OS anyway")


class CdromDisk(Disk):

    """
    Represents a CDROM disk that we can master according to our needs.
    """

    def __init__(self, path, tmpdir):
        self.mount = tempfile.mkdtemp(prefix='cdrom_virttest_', dir=tmpdir)
        self.tmpdir = tmpdir
        self.path = path
        clean_old_image(path)
        if not os.path.isdir(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))

    def _copy_virtio_drivers(self, virtio_floppy, cdrom_virtio):
        """
        Copy the virtio drivers from floppy or cdrom to install cdrom.

        1) If cdrom is available, mount the cdrom containing the virtio drivers
           and copy its contents to the root of the install cdrom
        2) If floppy is availabe while cdrom is not, mount the floppy and copy
           its contents to the root of the install cdrom
        """
        if cdrom_virtio:
            pwd = os.getcwd()
            mnt_pnt = tempfile.mkdtemp(prefix='cdrom_virtio_', dir=self.tmpdir)
            mount(cdrom_virtio, mnt_pnt, options='loop,ro', verbose=DEBUG)
            try:
                copytree(mnt_pnt, self.mount, ignore='*.vfd')
            finally:
                os.chdir(pwd)
                umount(None, mnt_pnt, verbose=DEBUG)
                os.rmdir(mnt_pnt)
        elif virtio_floppy:
            cmd = 'mcopy -s -o -n -i %s ::/* %s' % (virtio_floppy, self.mount)
            process.run(cmd, verbose=DEBUG)

    def setup_virtio_win2008(self, virtio_floppy, cdrom_virtio):
        """
        Setup the install cdrom with the virtio storage drivers, win2008 style.

        Win2008, Vista and 7 require people to point out the path to the drivers
        on the unattended file, so we just need to copy the drivers to the
        extra cdrom disk. Important to note that it's possible to specify
        drivers from a CDROM, so the floppy driver copy is optional.
        Process:

        1) Copy the virtio drivers on the virtio floppy/cdrom to the install
           cdrom if there is one available
        """
        if os.path.isfile(cdrom_virtio) or os.path.isfile(virtio_floppy):
            self._copy_virtio_drivers(virtio_floppy, cdrom_virtio)
        else:
            logging.debug(
                "No virtio floppy/cdrom present, not needed for this OS anyway")

    @error_context.context_aware
    def close(self):
        error_context.context(
            "Creating unattended install CD image %s" % self.path)
        g_cmd = ('mkisofs -o %s -max-iso9660-filenames '
                 '-relaxed-filenames -D --input-charset iso8859-1 '
                 '%s' % (self.path, self.mount))
        process.run(g_cmd, verbose=DEBUG)

        os.chmod(self.path, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP |
                 stat.S_IROTH | stat.S_IXOTH)
        cleanup(self.mount)
        logging.debug("unattended install CD image %s successfully created",
                      self.path)


class CdromInstallDisk(Disk):

    """
    Represents a install CDROM disk that we can master according to our needs.
    """

    def __init__(self, path, tmpdir, source_cdrom, extra_params):
        self.mount = tempfile.mkdtemp(prefix='cdrom_unattended_', dir=tmpdir)
        self.path = path
        self.extra_params = extra_params
        self.source_cdrom = source_cdrom
        cleanup(path)
        if not os.path.isdir(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        cp_cmd = ('cp -r %s/isolinux/ %s/' % (source_cdrom, self.mount))
        listdir = os.listdir(self.source_cdrom)
        for i in listdir:
            if i == 'isolinux':
                continue
            os.symlink(os.path.join(self.source_cdrom, i),
                       os.path.join(self.mount, i))
        process.run(cp_cmd)

    def get_answer_file_path(self, filename):
        return os.path.join(self.mount, 'isolinux', filename)

    @error_context.context_aware
    def close(self):
        error_context.context(
            "Creating unattended install CD image %s" % self.path)
        if os.path.exists(os.path.join(self.mount, 'isolinux')):
            # bootable cdrom
            f = open(os.path.join(self.mount, 'isolinux', 'isolinux.cfg'), 'w')
            f.write('default /isolinux/vmlinuz append initrd=/isolinux/'
                    'initrd.img %s\n' % self.extra_params)
            f.close()
            boot = '-b isolinux/isolinux.bin'
        else:
            # Not a bootable CDROM, using -kernel instead (eg.: arm64)
            boot = ''

        m_cmd = ('mkisofs -o %s %s -c isolinux/boot.cat -no-emul-boot '
                 '-boot-load-size 4 -boot-info-table -f -R -J -V -T %s'
                 % (self.path, boot, self.mount))
        process.run(m_cmd)
        os.chmod(self.path, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP |
                 stat.S_IROTH | stat.S_IXOTH)
        cleanup(self.mount)
        cleanup(self.source_cdrom)
        logging.debug("unattended install CD image %s successfully created",
                      self.path)


class GuestFSModiDisk(object):

    """
    class of guest disk using guestfs lib to do some operation(like read/write)
    on guest disk:
    """

    def __init__(self, disk, backend='direct'):
        """
        :params disk: target disk image.
        :params backend: let libguestfs creates/connects to backend daemon
                         by starting qemu directly, or using libvirt to manage
                         an appliance, running User-Mode Linux, or connecting
                         to an already running daemon.
                         'direct', 'appliance', 'libvirt', 'libvirt:null',
                         'libvirt:URI', 'uml', 'unix:path'.
        """
        try:
            import guestfs
        except ImportError:
            from virttest.utils_package import package_install
            if not package_install("python*-libguestfs"):
                raise exceptions.TestSkipError('We need python-libguestfs (or '
                                               'the equivalent for your '
                                               'distro) for this particular '
                                               'feature (modifying guest '
                                               'files with libguestfs)')
            try:
                import guestfs
            except ImportError:
                raise exceptions.TestError("Couldn't import guestfs")

        self.g = guestfs.GuestFS()
        self.disk = disk
        self.g.add_drive(disk)
        self.g.set_backend(backend)
        libvirtd = SpecificServiceManager("libvirtd")
        libvirtd_status = libvirtd.status()
        if libvirtd_status is None:
            raise exceptions.TestError('libvirtd: service not found')
        if (not libvirtd_status) and (not libvirtd.start()):
            raise exceptions.TestError('libvirtd: failed to start')
        logging.debug("Launch the disk %s, wait..." % self.disk)
        self.g.launch()

    def os_inspects(self):
        self.roots = self.g.inspect_os()
        if self.roots:
            return self.roots
        else:
            return None

    def mounts(self):
        return self.g.mounts()

    def mount_all(self):
        def compare(a, b):
            if len(a[0]) > len(b[0]):
                return 1
            elif len(a[0]) == len(b[0]):
                return 0
            else:
                return -1

        roots = self.os_inspects()
        if roots:
            for root in roots:
                mps = self.g.inspect_get_mountpoints(root)
                mps.sort(key=cmp_to_key(compare))
                for mp_dev in mps:
                    try:
                        msg = "Mount dev '%s' partitions '%s' to '%s'"
                        logging.info(msg % (root, mp_dev[1], mp_dev[0]))
                        self.g.mount(mp_dev[1], mp_dev[0])
                    except RuntimeError as err_msg:
                        logging.info("%s (ignored)" % err_msg)
        else:
            raise exceptions.TestError(
                "inspect_vm: no operating systems found")

    def umount_all(self):
        logging.debug("Umount all device partitions")
        if self.mounts():
            self.g.umount_all()

    def read_file(self, file_name):
        """
        read file from the guest disk, return the content of the file

        :param file_name: the file you want to read.
        """

        try:
            self.mount_all()
            o = self.g.cat(file_name)
            if o:
                return o
            else:
                err_msg = "Can't read file '%s', check is it exist?"
                raise exceptions.TestError(err_msg % file_name)
        finally:
            self.umount_all()

    def write_to_image_file(self, file_name, content, w_append=False):
        """
        Write content to the file on the guest disk.

        When using this method all the original content will be overriding.
        if you don't hope your original data be override set ``w_append=True``.

        :param file_name: the file you want to write
        :param content: the content you want to write.
        :param w_append: append the content or override
        """

        try:
            try:
                self.mount_all()
                if w_append:
                    self.g.write_append(file_name, content)
                else:
                    self.g.write(file_name, content)
            except Exception:
                raise exceptions.TestError("write '%s' to file '%s' error!"
                                           % (content, file_name))
        finally:
            self.umount_all()

    def replace_image_file_content(self, file_name, find_con, rep_con):
        """
        replace file content matches in the file with rep_con.
        support using Regular expression

        :param file_name: the file you want to replace
        :param find_con: the original content you want to replace.
        :param rep_con: the replace content you want.
        """

        try:
            self.mount_all()
            file_content = self.g.cat(file_name)
            if file_content:
                file_content_after_replace = re.sub(find_con, rep_con,
                                                    file_content)
                if file_content != file_content_after_replace:
                    self.g.write(file_name, file_content_after_replace)
            else:
                err_msg = "Can't read file '%s', check is it exist?"
                raise exceptions.TestError(err_msg % file_name)
        finally:
            self.umount_all()

    def close(self):
        """
        Explicitly close the guestfs handle.
        """
        if self.g:
            self.g.close()
